# -----------------------------------------------------------------------------
# Copyright (c) 2023, 2024, Oracle and/or its affiliates.
#
# This software is dual-licensed to you under the Universal Permissive License
# (UPL) 1.0 as shown at https://oss.oracle.com/licenses/upl and Apache License
# 2.0 as shown at http://www.apache.org/licenses/LICENSE-2.0. You may choose
# either license.
#
# If you elect to accept the software under the Apache License, Version 2.0,
# the following applies:
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# -----------------------------------------------------------------------------

"""
5900 - Module for testing DML returning clauses with asyncio
"""

import datetime
import unittest

import oracledb
import test_env


@unittest.skipUnless(
    test_env.get_is_thin(), "asyncio not supported in thick mode"
)
class TestCase(test_env.BaseAsyncTestCase):
    async def test_5900(self):
        "5900 - test insert (single row) with DML returning"
        await self.cursor.execute("truncate table TestTempTable")
        int_val = 5
        str_val = "A test string"
        int_var = self.cursor.var(oracledb.NUMBER)
        str_var = self.cursor.var(str)
        await self.cursor.execute(
            """
            insert into TestTempTable (IntCol, StringCol1)
            values (:int_val, :str_val)
            returning IntCol, StringCol1 into :int_var, :str_var
            """,
            int_val=int_val,
            str_val=str_val,
            int_var=int_var,
            str_var=str_var,
        )
        self.assertEqual(int_var.values, [[int_val]])
        self.assertEqual(str_var.values, [[str_val]])

    async def test_5901(self):
        "5901 - test insert (multiple rows) with DML returning"
        await self.cursor.execute("truncate table TestTempTable")
        int_values = [5, 8, 17, 24, 6]
        str_values = ["Test 5", "Test 8", "Test 17", "Test 24", "Test 6"]
        int_var = self.cursor.var(oracledb.NUMBER, arraysize=len(int_values))
        str_var = self.cursor.var(str, arraysize=len(int_values))
        self.cursor.setinputsizes(None, None, int_var, str_var)
        data = list(zip(int_values, str_values))
        await self.cursor.executemany(
            """
            insert into TestTempTable (IntCol, StringCol1)
            values (:int_val, :str_val)
            returning IntCol, StringCol1 into :int_var, :str_var
            """,
            data,
        )
        self.assertEqual(int_var.values, [[v] for v in int_values])
        self.assertEqual(str_var.values, [[v] for v in str_values])

    async def test_5902(self):
        "5902 - test insert with DML returning into too small a variable"
        await self.cursor.execute("truncate table TestTempTable")
        int_val = 6
        str_val = "A different test string"
        int_var = self.cursor.var(oracledb.NUMBER)
        str_var = self.cursor.var(str, 2)
        parameters = dict(
            int_val=int_val, str_val=str_val, int_var=int_var, str_var=str_var
        )
        with self.assertRaisesFullCode("DPY-4002", "DPI-1037"):
            await self.cursor.execute(
                """
                insert into TestTempTable (IntCol, StringCol1)
                values (:int_val, :str_val)
                returning IntCol, StringCol1 into :int_var, :str_var
                """,
                parameters,
            )

    async def test_5903(self):
        "5903 - test update single row with DML returning"
        int_val = 7
        str_val = "The updated value of the string"
        await self.cursor.execute("truncate table TestTempTable")
        await self.cursor.execute(
            """
            insert into TestTempTable (IntCol, StringCol1)
            values (:1, :2)
            """,
            (int_val, "The initial value of the string"),
        )
        int_var = self.cursor.var(oracledb.NUMBER)
        str_var = self.cursor.var(str)
        await self.cursor.execute(
            """
            update TestTempTable set
                StringCol1 = :str_val
            where IntCol = :int_val
            returning IntCol, StringCol1 into :int_var, :str_var
            """,
            int_val=int_val,
            str_val=str_val,
            int_var=int_var,
            str_var=str_var,
        )
        self.assertEqual(int_var.values, [[int_val]])
        self.assertEqual(str_var.values, [[str_val]])

    async def test_5904(self):
        "5904 - test update no rows with DML returning"
        int_val = 8
        str_val = "The updated value of the string"
        await self.cursor.execute("truncate table TestTempTable")
        await self.cursor.execute(
            """
            insert into TestTempTable (IntCol, StringCol1)
            values (:1, :2)
            """,
            (int_val, "The initial value of the string"),
        )
        int_var = self.cursor.var(oracledb.NUMBER)
        str_var = self.cursor.var(str)
        await self.cursor.execute(
            """
            update TestTempTable set
                StringCol1 = :str_val
            where IntCol = :int_val
            returning IntCol, StringCol1 into :int_var, :str_var
            """,
            int_val=int_val + 1,
            str_val=str_val,
            int_var=int_var,
            str_var=str_var,
        )
        self.assertEqual(int_var.values, [[]])
        self.assertEqual(str_var.values, [[]])
        self.assertEqual(int_var.getvalue(), [])
        self.assertEqual(str_var.getvalue(), [])

    async def test_5905(self):
        "5905 - test update multiple rows with DML returning"
        await self.cursor.execute("truncate table TestTempTable")
        for i in (8, 9, 10):
            await self.cursor.execute(
                """
                insert into TestTempTable (IntCol, StringCol1) values (:1, :2)
                """,
                (i, f"The initial value of string {i}"),
            )
        int_var = self.cursor.var(oracledb.NUMBER)
        str_var = self.cursor.var(str)
        await self.cursor.execute(
            """
            update TestTempTable set
                IntCol = IntCol + 15,
                StringCol1 = 'The final value of string ' || to_char(IntCol)
            returning IntCol, StringCol1 into :int_var, :str_var
            """,
            int_var=int_var,
            str_var=str_var,
        )
        self.assertEqual(self.cursor.rowcount, 3)
        self.assertEqual(int_var.values, [[23, 24, 25]])
        expected_values = [
            [
                "The final value of string 8",
                "The final value of string 9",
                "The final value of string 10",
            ]
        ]
        self.assertEqual(str_var.values, expected_values)

    async def test_5906(self):
        "5906 - test update multiple rows with DML returning (executemany)"
        data = [(i, f"The initial value of string {i}") for i in range(1, 11)]
        await self.cursor.execute("truncate table TestTempTable")
        await self.cursor.executemany(
            "insert into TestTempTable (IntCol, StringCol1) values (:1, :2)",
            data,
        )
        int_var = self.cursor.var(oracledb.NUMBER, arraysize=3)
        str_var = self.cursor.var(str, arraysize=3)
        self.cursor.setinputsizes(None, int_var, str_var)
        await self.cursor.executemany(
            """
            update TestTempTable set
                IntCol = IntCol + 25,
                StringCol1 = 'Updated value of string ' || to_char(IntCol)
            where IntCol < :inVal
            returning IntCol, StringCol1 into :int_var, :str_var
            """,
            [[3], [8], [11]],
        )
        expected_values = [[26, 27], [28, 29, 30, 31, 32], [33, 34, 35]]
        self.assertEqual(int_var.values, expected_values)
        expected_values = [
            ["Updated value of string 1", "Updated value of string 2"],
            [
                "Updated value of string 3",
                "Updated value of string 4",
                "Updated value of string 5",
                "Updated value of string 6",
                "Updated value of string 7",
            ],
            [
                "Updated value of string 8",
                "Updated value of string 9",
                "Updated value of string 10",
            ],
        ]
        self.assertEqual(str_var.values, expected_values)

    async def test_5907(self):
        "5907 - test inserting an object with DML returning"
        type_obj = await self.conn.gettype("UDT_OBJECT")
        string_value = "The string that will be verified"
        obj = type_obj.newobject()
        obj.STRINGVALUE = string_value
        out_var = self.cursor.var(type_obj)
        await self.cursor.execute(
            """
            insert into TestObjects (IntCol, ObjectCol)
            values (4, :obj)returning ObjectCol into :outObj
            """,
            obj=obj,
            outObj=out_var,
        )
        (result,) = out_var.getvalue()
        self.assertEqual(result.STRINGVALUE, string_value)
        await self.conn.rollback()

    async def test_5908(self):
        "5908 - test inserting a row and returning a rowid"
        await self.cursor.execute("truncate table TestTempTable")
        var = self.cursor.var(oracledb.ROWID)
        await self.cursor.execute(
            """
            insert into TestTempTable (IntCol, StringCol1)
            values (278, 'String 278')
            returning rowid into :1
            """,
            [var],
        )
        (rowid,) = var.getvalue()
        await self.cursor.execute(
            """
            select IntCol, StringCol1
            from TestTempTable
            where rowid = :1
            """,
            [rowid],
        )
        self.assertEqual(await self.cursor.fetchall(), [(278, "String 278")])

    async def test_5909(self):
        "5909 - test inserting with a REF cursor and returning a rowid"
        await self.cursor.execute("truncate table TestTempTable")
        var = self.cursor.var(oracledb.ROWID)
        in_cursor = self.conn.cursor()
        await in_cursor.execute(
            """
            select StringCol
            from TestStrings
            where IntCol >= 5
            order by IntCol
            """
        )
        await self.cursor.execute(
            """
            insert into TestTempTable (IntCol, StringCol1)
            values (187, pkg_TestRefCursors.TestInCursor(:1))
            returning rowid into :2
            """,
            (in_cursor, var),
        )
        (rowid,) = var.getvalue()
        await self.cursor.execute(
            """
            select IntCol, StringCol1
            from TestTempTable
            where rowid = :1
            """,
            [rowid],
        )
        self.assertEqual(
            await self.cursor.fetchall(), [(187, "String 7 (Modified)")]
        )

    async def test_5910(self):
        "5910 - test delete returning decreasing number of rows"
        data = [(i, f"Test String {i}") for i in range(1, 11)]
        await self.cursor.execute("truncate table TestTempTable")
        await self.cursor.executemany(
            """
            insert into TestTempTable (IntCol, StringCol1)
            values (:1, :2)
            """,
            data,
        )
        results = []
        int_var = self.cursor.var(int)
        self.cursor.setinputsizes(None, int_var)
        for int_val in (5, 8, 10):
            await self.cursor.execute(
                """
                delete from TestTempTable
                where IntCol < :1
                returning IntCol into :2
                """,
                [int_val],
            )
            results.append(int_var.getvalue())
        self.assertEqual(results, [[1, 2, 3, 4], [5, 6, 7], [8, 9]])

    async def test_5911(self):
        "5911 - test delete returning no rows after returning many rows"
        data = [(i, f"Test String {i}") for i in range(1, 11)]
        await self.cursor.execute("truncate table TestTempTable")
        await self.cursor.executemany(
            """
            insert into TestTempTable (IntCol, StringCol1)
            values (:1, :2)
            """,
            data,
        )
        int_var = self.cursor.var(int)
        await self.cursor.execute(
            """
            delete from TestTempTable
            where IntCol < :1
            returning IntCol into :2
            """,
            [5, int_var],
        )
        self.assertEqual(int_var.getvalue(), [1, 2, 3, 4])
        await self.cursor.execute(None, [4, int_var])
        self.assertEqual(int_var.getvalue(), [])

    async def test_5912(self):
        "5912 - test DML returning when an error occurs"
        await self.cursor.execute("truncate table TestTempTable")
        int_val = 7
        str_val = "A" * 401
        int_var = self.cursor.var(oracledb.NUMBER)
        str_var = self.cursor.var(str)
        sql = """
                insert into TestTempTable (IntCol, StringCol1)
                values (:int_val, :str_val)
                returning IntCol, StringCol1 into :int_var, :str_var"""
        parameters = dict(
            int_val=int_val, str_val=str_val, int_var=int_var, str_var=str_var
        )
        with self.assertRaisesFullCode("ORA-12899"):
            await self.cursor.execute(sql, parameters)

    async def test_5913(self):
        "5913 - test DML returning with no input variables, multiple iters"
        await self.cursor.execute("truncate table TestTempTable")
        sql = """
                insert into TestTempTable (IntCol)
                values ((select count(*) + 1 from TestTempTable))
                returning IntCol into :1"""
        var = self.cursor.var(int)
        await self.cursor.execute(sql, [var])
        self.assertEqual(var.getvalue(), [1])
        await self.cursor.execute(sql, [var])
        self.assertEqual(var.getvalue(), [2])

    async def test_5914(self):
        "5914 - test DML returning with a quoted bind name"
        sql = """
                insert into TestTempTable (IntCol, StringCol1)
                values (:int_val, :str_val)
                returning IntCol, StringCol1 into :"_val1" , :"VaL_2" """
        await self.cursor.parse(sql)
        expected_bind_names = ["INT_VAL", "STR_VAL", "_val1", "VaL_2"]
        self.assertEqual(self.cursor.bindnames(), expected_bind_names)

    async def test_5915(self):
        "5915 - test DML returning with an invalid bind name"
        sql = """
                insert into TestTempTable (IntCol)
                values (:int_val)
                returning IntCol, StringCol1 into :ROWID"""
        with self.assertRaisesFullCode("ORA-01745"):
            await self.cursor.parse(sql)

    async def test_5916(self):
        "5916 - test DML returning with input bind variable data"
        await self.cursor.execute("truncate table TestTempTable")
        out_var = self.cursor.var(int)
        await self.cursor.execute(
            """
            insert into TestTempTable (IntCol)
            values (:int_val)
            returning IntCol + :add_val into :out_val
            """,
            int_val=5,
            add_val=18,
            out_val=out_var,
        )
        await self.conn.commit()
        self.assertEqual(out_var.getvalue(), [23])

    async def test_5917(self):
        "5917 - test DML returning with LOBs and an output converter"
        await self.cursor.execute("truncate table TestCLOBs")
        out_var = self.cursor.var(
            oracledb.DB_TYPE_CLOB, outconverter=lambda value: value.read()
        )
        lob_value = "A short CLOB - 1618"
        await self.cursor.execute(
            """
            insert into TestCLOBs (IntCol, ClobCol)
            values (1, :in_val)
            returning CLOBCol into :out_val
            """,
            in_val=lob_value,
            out_val=out_var,
        )
        await self.conn.commit()
        self.assertEqual(out_var.getvalue(), [lob_value])

    async def test_5918(self):
        "5918 - test DML returning with CLOB converted to LONG"
        await self.cursor.execute("truncate table TestCLOBs")
        out_var = self.cursor.var(oracledb.DB_TYPE_LONG)
        lob_value = "A short CLOB - 1619"
        await self.cursor.execute(
            """
            insert into TestCLOBs
            (IntCol, ClobCol)
            values (1, :in_val)
            returning CLOBCol into :out_val
            """,
            in_val=lob_value,
            out_val=out_var,
        )
        await self.conn.commit()
        self.assertEqual(out_var.getvalue(), [lob_value])

    async def test_5919(self):
        "5919 - test dml returning with an index organized table"
        await self.cursor.execute("truncate table TestUniversalRowids")
        rowid_var = self.cursor.var(oracledb.ROWID)
        data = (1, "ABC", datetime.datetime(2017, 4, 11), rowid_var)
        sql = """
                insert into TestUniversalRowids values (:1, :2, :3)
                returning rowid into :4"""
        await self.cursor.execute(sql, data)
        (rowid_value,) = rowid_var.getvalue()
        await self.cursor.execute(
            """
            select *
            from TestUniversalRowids
            where rowid = :1
            """,
            [rowid_value],
        )
        (row,) = await self.cursor.fetchall()
        self.assertEqual(row, data[:3])

    async def test_5920(self):
        "5920 - test plsql returning rowids with index organized table"
        await self.cursor.execute("truncate table TestUniversalRowids")
        rowid_var = self.cursor.var(oracledb.ROWID)
        data = (1, "ABC", datetime.datetime(2017, 4, 11), rowid_var)
        await self.cursor.execute(
            """
            begin
                insert into TestUniversalRowids values (:1, :2, :3)
                returning rowid into :4;
            end;
            """,
            data,
        )
        rowid_value = rowid_var.values[0]
        await self.cursor.execute(
            """
            select *
            from TestUniversalRowids
            where rowid = :1
            """,
            [rowid_value],
        )
        (row,) = await self.cursor.fetchall()
        self.assertEqual(row, data[:3])

    async def test_5921(self):
        "5921 - parse DML returning with no spaces"
        await self.cursor.execute("truncate table TestTempTable")
        sql = (
            "insert into TestTempTable (IntCol) values (:in_val)"
            "returning(IntCol)into :out_val"
        )
        out_val = self.cursor.var(int, arraysize=5)
        await self.cursor.execute(sql, in_val=25, out_val=out_val)
        self.assertEqual(out_val.getvalue(), [25])


if __name__ == "__main__":
    test_env.run_test_cases()



#------------------------------------------------------------------------------
# Copyright (c) 2020, 2024, Oracle and/or its affiliates.
#
# This software is dual-licensed to you under the Universal Permissive License
# (UPL) 1.0 as shown at https://oss.oracle.com/licenses/upl and Apache License
# 2.0 as shown at http://www.apache.org/licenses/LICENSE-2.0. You may choose
# either license.
#
# If you elect to accept the software under the Apache License, Version 2.0,
# the following applies:
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#------------------------------------------------------------------------------

#------------------------------------------------------------------------------
# base_impl.pxd
#
# Cython definition file defining the base classes from which the thick and
# thin implementations derive their classes.
#------------------------------------------------------------------------------

# cython: language_level=3

from libc.stdint cimport int8_t, int16_t, int32_t, int64_t
from libc.stdint cimport uint8_t, uint16_t, uint32_t, uint64_t
from cpython cimport array

ctypedef unsigned char char_type

cdef enum:
    NUM_TYPE_FLOAT = 0
    NUM_TYPE_INT = 1
    NUM_TYPE_DECIMAL = 2
    NUM_TYPE_STR = 3

cdef enum:
    DB_TYPE_NUM_MIN = 2000
    DB_TYPE_NUM_MAX = 2034

    DB_TYPE_NUM_BFILE = 2020
    DB_TYPE_NUM_BINARY_DOUBLE = 2008
    DB_TYPE_NUM_BINARY_FLOAT = 2007
    DB_TYPE_NUM_BINARY_INTEGER = 2009
    DB_TYPE_NUM_BLOB = 2019
    DB_TYPE_NUM_BOOLEAN = 2022
    DB_TYPE_NUM_CHAR = 2003
    DB_TYPE_NUM_CLOB = 2017
    DB_TYPE_NUM_CURSOR = 2021
    DB_TYPE_NUM_DATE = 2011
    DB_TYPE_NUM_INTERVAL_DS = 2015
    DB_TYPE_NUM_INTERVAL_YM = 2016
    DB_TYPE_NUM_JSON = 2027
    DB_TYPE_NUM_LONG_NVARCHAR = 2031
    DB_TYPE_NUM_LONG_RAW = 2025
    DB_TYPE_NUM_LONG_VARCHAR = 2024
    DB_TYPE_NUM_NCHAR = 2004
    DB_TYPE_NUM_NCLOB = 2018
    DB_TYPE_NUM_NUMBER = 2010
    DB_TYPE_NUM_NVARCHAR = 2002
    DB_TYPE_NUM_OBJECT = 2023
    DB_TYPE_NUM_RAW = 2006
    DB_TYPE_NUM_ROWID = 2005
    DB_TYPE_NUM_TIMESTAMP = 2012
    DB_TYPE_NUM_TIMESTAMP_LTZ = 2014
    DB_TYPE_NUM_TIMESTAMP_TZ = 2013
    DB_TYPE_NUM_UNKNOWN = 0
    DB_TYPE_NUM_UROWID = 2030
    DB_TYPE_NUM_VARCHAR = 2001
    DB_TYPE_NUM_VECTOR = 2033
    DB_TYPE_NUM_XMLTYPE = 2032

cdef enum:
    NATIVE_TYPE_NUM_BOOLEAN = 3011
    NATIVE_TYPE_NUM_BYTES = 3004
    NATIVE_TYPE_NUM_DOUBLE = 3003
    NATIVE_TYPE_NUM_FLOAT = 3002
    NATIVE_TYPE_NUM_INTERVAL_DS = 3006
    NATIVE_TYPE_NUM_INTERVAL_YM = 3007
    NATIVE_TYPE_NUM_INT64 = 3000
    NATIVE_TYPE_NUM_JSON = 3013
    NATIVE_TYPE_NUM_LOB = 3008
    NATIVE_TYPE_NUM_OBJECT = 3009
    NATIVE_TYPE_NUM_ROWID = 3012
    NATIVE_TYPE_NUM_STMT = 3010
    NATIVE_TYPE_NUM_TIMESTAMP = 3005
    NATIVE_TYPE_NUM_VECTOR = 3017

cdef enum:
    CS_FORM_IMPLICIT = 1
    CS_FORM_NCHAR = 2

cdef enum:
    BYTE_ORDER_LSB = 1
    BYTE_ORDER_MSB = 2

cdef enum:
    TNS_LONG_LENGTH_INDICATOR = 254
    TNS_NULL_LENGTH_INDICATOR = 255

cpdef enum:
    AUTH_MODE_DEFAULT = 0
    AUTH_MODE_PRELIM = 0x00000008
    AUTH_MODE_SYSASM = 0x00008000
    AUTH_MODE_SYSBKP = 0x00020000
    AUTH_MODE_SYSDBA = 0x00000002
    AUTH_MODE_SYSDGD = 0x00040000
    AUTH_MODE_SYSKMT = 0x00080000
    AUTH_MODE_SYSOPER = 0x00000004
    AUTH_MODE_SYSRAC = 0x00100000

cpdef enum:
    PIPELINE_OP_TYPE_CALL_FUNC = 1
    PIPELINE_OP_TYPE_CALL_PROC = 2
    PIPELINE_OP_TYPE_COMMIT = 3
    PIPELINE_OP_TYPE_EXECUTE = 4
    PIPELINE_OP_TYPE_EXECUTE_MANY = 5
    PIPELINE_OP_TYPE_FETCH_ALL = 6
    PIPELINE_OP_TYPE_FETCH_MANY = 7
    PIPELINE_OP_TYPE_FETCH_ONE = 8

cpdef enum:
    POOL_GETMODE_WAIT = 0
    POOL_GETMODE_NOWAIT = 1
    POOL_GETMODE_FORCEGET = 2
    POOL_GETMODE_TIMEDWAIT = 3

cpdef enum:
    PURITY_DEFAULT = 0
    PURITY_NEW = 1
    PURITY_SELF = 2

cpdef enum:
    VECTOR_FORMAT_BINARY = 5
    VECTOR_FORMAT_FLOAT32 = 2
    VECTOR_FORMAT_FLOAT64 = 3
    VECTOR_FORMAT_INT8 = 4

cdef type PY_TYPE_ASYNC_CURSOR
cdef type PY_TYPE_ASYNC_LOB
cdef type PY_TYPE_BOOL
cdef type PY_TYPE_CURSOR
cdef type PY_TYPE_DATE
cdef type PY_TYPE_DATETIME
cdef type PY_TYPE_DECIMAL
cdef type PY_TYPE_DB_OBJECT
cdef type PY_TYPE_DB_OBJECT_TYPE
cdef type PY_TYPE_FETCHINFO
cdef type PY_TYPE_JSON_ID
cdef type PY_TYPE_INTERVAL_YM
cdef type PY_TYPE_LOB
cdef type PY_TYPE_MESSAGE
cdef type PY_TYPE_MESSAGE_QUERY
cdef type PY_TYPE_MESSAGE_ROW
cdef type PY_TYPE_MESSAGE_TABLE
cdef type PY_TYPE_TIMEDELTA
cdef type PY_TYPE_VAR

cdef str DRIVER_NAME
cdef str DRIVER_VERSION
cdef str DRIVER_INSTALLATION_URL

cdef const char* ENCODING_UTF8
cdef const char* ENCODING_UTF16

cdef class ApiType:
    cdef:
        readonly str name
        tuple dbtypes


cdef class DbType:
    cdef:
        readonly uint32_t num
        readonly str name
        readonly uint32_t default_size
        uint32_t _native_num
        uint32_t _buffer_size_factor
        str _ora_name
        uint8_t _ora_type_num
        uint8_t _csfrm

    @staticmethod
    cdef DbType _from_num(uint32_t num)

    @staticmethod
    cdef DbType _from_ora_name(str name)

    @staticmethod
    cdef DbType _from_ora_type_and_csfrm(uint8_t ora_type_num, uint8_t csfrm)


cdef class DefaultsImpl:
    cdef:
        public uint32_t arraysize
        public str config_dir
        public bint fetch_lobs
        public bint fetch_decimals
        public uint32_t prefetchrows
        public uint32_t stmtcachesize

cdef DefaultsImpl C_DEFAULTS


cdef class Buffer:
    cdef:
        ssize_t _max_size, _size, _pos
        char_type[:] _data_view
        char_type *_data
        bytearray _data_obj

    cdef int _get_int_length_and_sign(self, uint8_t *length,
                                      bint *is_negative,
                                      uint8_t max_length) except -1
    cdef const char_type* _get_raw(self, ssize_t num_bytes) except NULL
    cdef int _initialize(self, ssize_t max_size=*) except -1
    cdef int _populate_from_bytes(self, bytes data) except -1
    cdef int _read_raw_bytes_and_length(self, const char_type **ptr,
                                        ssize_t *num_bytes) except -1
    cdef int _resize(self, ssize_t new_max_size) except -1
    cdef int _skip_int(self, uint8_t max_length, bint *is_negative) except -1
    cdef uint64_t _unpack_int(self, const char_type *ptr, uint8_t length)
    cdef int _write_more_data(self, ssize_t num_bytes_available,
                              ssize_t num_bytes_wanted) except -1
    cdef int _write_raw_bytes_and_length(self, const char_type *ptr,
                                         ssize_t num_bytes) except -1
    cdef inline ssize_t bytes_left(self)
    cdef int parse_binary_double(self, const uint8_t* ptr,
                                 double *double_ptr) except -1
    cdef int parse_binary_float(self, const uint8_t* ptr,
                                float *float_ptr) except -1
    cdef object parse_date(self, const uint8_t* ptr, ssize_t num_bytes)
    cdef object parse_interval_ds(self, const uint8_t* ptr)
    cdef object parse_interval_ym(self, const uint8_t* ptr)
    cdef object parse_oracle_number(self, const uint8_t* ptr,
                                    ssize_t num_bytes, int preferred_num_type)
    cdef object read_binary_double(self)
    cdef object read_binary_float(self)
    cdef object read_binary_integer(self)
    cdef object read_bool(self)
    cdef object read_bytes(self)
    cdef object read_date(self)
    cdef object read_interval_ds(self)
    cdef object read_interval_ym(self)
    cdef int read_int32(self, int32_t *value, int byte_order=*) except -1
    cdef object read_oracle_number(self, int preferred_num_type)
    cdef const char_type* read_raw_bytes(self, ssize_t num_bytes) except NULL
    cdef int read_raw_bytes_and_length(self, const char_type **ptr,
                                       ssize_t *num_bytes) except -1
    cdef int read_sb1(self, int8_t *value) except -1
    cdef int read_sb2(self, int16_t *value) except -1
    cdef int read_sb4(self, int32_t *value) except -1
    cdef int read_sb8(self, int64_t *value) except -1
    cdef bytes read_null_terminated_bytes(self)
    cdef object read_str(self, int csfrm, const char* encoding_errors=*)
    cdef int read_ub1(self, uint8_t *value) except -1
    cdef int read_ub2(self, uint16_t *value) except -1
    cdef int read_ub4(self, uint32_t *value) except -1
    cdef int read_ub8(self, uint64_t *value) except -1
    cdef int read_uint16(self, uint16_t *value, int byte_order=*) except -1
    cdef int read_uint32(self, uint32_t *value, int byte_order=*) except -1
    cdef int skip_raw_bytes(self, ssize_t num_bytes) except -1
    cdef inline int skip_sb4(self) except -1
    cdef inline void skip_to(self, ssize_t pos)
    cdef inline int skip_ub1(self) except -1
    cdef inline int skip_ub2(self) except -1
    cdef inline int skip_ub4(self) except -1
    cdef inline int skip_ub8(self) except -1
    cdef int write_binary_double(self, double value,
                                 bint write_length=*) except -1
    cdef int write_binary_float(self, float value,
                                bint write_length=*) except -1
    cdef int write_bool(self, bint value) except -1
    cdef int write_bytes(self, bytes value) except -1
    cdef int write_bytes_with_length(self, bytes value) except -1
    cdef int write_interval_ds(self, object value,
                               bint write_length=*) except -1
    cdef int write_interval_ym(self, object value,
                               bint write_length=*) except -1
    cdef int write_oracle_date(self, object value, uint8_t length,
                               bint write_length=*) except -1
    cdef int write_oracle_number(self, bytes num_bytes) except -1
    cdef int write_raw(self, const char_type *data, ssize_t length) except -1
    cdef int write_str(self, str value) except -1
    cdef int write_uint8(self, uint8_t value) except -1
    cdef int write_uint16(self, uint16_t value, int byte_order=*) except -1
    cdef int write_uint32(self, uint32_t value, int byte_order=*) except -1
    cdef int write_uint64(self, uint64_t value, byte_order=*) except -1
    cdef int write_ub2(self, uint16_t value) except -1
    cdef int write_ub4(self, uint32_t value) except -1
    cdef int write_ub8(self, uint64_t value) except -1


cdef class GrowableBuffer(Buffer):

    cdef int _reserve_space(self, ssize_t num_bytes) except -1
    cdef int _write_more_data(self, ssize_t num_bytes_available,
                              ssize_t num_bytes_wanted) except -1


cdef class OsonDecoder(Buffer):

    cdef:
        uint16_t primary_flags, secondary_flags
        ssize_t field_id_length
        ssize_t tree_seg_pos
        list field_names
        uint8_t version
        bint relative_offsets

    cdef object _decode_container_node(self, uint8_t node_type)
    cdef object _decode_node(self)
    cdef list _get_long_field_names(self, uint32_t num_fields,
                                    ssize_t offsets_size,
                                    uint32_t field_names_seg_size)
    cdef int _get_num_children(self, uint8_t node_type, uint32_t* num_children,
                               bint* is_shared) except -1
    cdef int _get_offset(self, uint8_t node_type, uint32_t* offset) except -1
    cdef list _get_short_field_names(self, uint32_t num_fields,
                                     ssize_t offsets_size,
                                     uint32_t field_names_seg_size)
    cdef object decode(self, bytes data)


cdef class OsonFieldName:

    cdef:
        str name
        bytes name_bytes
        ssize_t name_bytes_len
        uint32_t hash_id
        uint32_t offset
        uint32_t field_id

    cdef int _calc_hash_id(self) except -1
    @staticmethod
    cdef OsonFieldName create(str name, ssize_t max_fname_size)


cdef class OsonFieldNamesSegment(GrowableBuffer):

    cdef:
        uint32_t num_field_names
        list field_names

    cdef int add_name(self, OsonFieldName field_name) except -1
    @staticmethod
    cdef OsonFieldNamesSegment create()
    cdef int process_field_names(self, ssize_t field_id_offset) except -1


cdef class OsonEncoder(GrowableBuffer):

    cdef:
        OsonFieldNamesSegment short_fnames_seg
        OsonFieldNamesSegment long_fnames_seg
        uint32_t num_field_names
        ssize_t max_fname_size
        dict field_names_dict
        uint8_t field_id_size

    cdef int _add_field_name(self, str name) except -1
    cdef int _determine_flags(self, object value, uint16_t *flags) except -1
    cdef int _examine_node(self, object value) except -1
    cdef int _write_extended_header(self) except -1
    cdef int _write_fnames_seg(self, OsonFieldNamesSegment seg) except -1
    cdef int encode(self, object value, ssize_t max_fname_size) except -1


cdef class VectorDecoder(Buffer):

    cdef object decode(self, bytes data)


cdef class VectorEncoder(GrowableBuffer):

    cdef int encode(self, array.array value) except -1


cdef class ConnectParamsNode:
    cdef:
        public bint source_route
        public bint load_balance
        public bint failover
        public bint must_have_children
        public list children
        public list active_children

    cdef int _copy(self, ConnectParamsNode source) except -1
    cdef int _set_active_children(self) except -1


cdef class Address(ConnectParamsNode):
    cdef:
        public str host
        public uint32_t port
        public str protocol
        public str https_proxy
        public uint32_t https_proxy_port

    cdef str build_connect_string(self)
    cdef int set_protocol(self, str value) except -1


cdef class AddressList(ConnectParamsNode):

    cdef bint _uses_tcps(self)
    cdef str build_connect_string(self)


cdef class Description(ConnectParamsNode):
    cdef:
        public uint32_t expire_time
        public uint32_t retry_count
        public uint32_t retry_delay
        public uint32_t sdu
        public double tcp_connect_timeout
        public str service_name
        public str server_type
        public str sid
        public str cclass
        public str connection_id_prefix
        public str pool_boundary
        public uint32_t purity
        public bint ssl_server_dn_match
        public bint use_tcp_fast_open
        public str ssl_server_cert_dn
        public object ssl_version
        public str wallet_location
        str connection_id

    cdef str _build_duration_str(self, double value)
    cdef str build_connect_string(self, str cid=*)
    cdef int set_server_type(self, str value) except -1


cdef class DescriptionList(ConnectParamsNode):

    cdef str build_connect_string(self)
    cdef list get_addresses(self)


cdef class ConnectParamsImpl:
    cdef:
        public str config_dir
        public str user
        public str proxy_user
        public bint events
        public bint externalauth
        public uint32_t mode
        public str edition
        public list appcontext
        public str tag
        public bint matchanytag
        public list shardingkey
        public list supershardingkey
        public uint32_t stmtcachesize
        public bint disable_oob
        public object ssl_context
        public DescriptionList description_list
        uint64_t _external_handle
        public str debug_jdwp
        object access_token_callback
        object access_token_expires
        Description _default_description
        Address _default_address
        bytearray _password
        bytearray _password_obfuscator
        bytearray _new_password
        bytearray _new_password_obfuscator
        bytearray _wallet_password
        bytearray _wallet_password_obfuscator
        bytearray _token
        bytearray _token_obfuscator
        bytearray _private_key
        bytearray _private_key_obfuscator

    cdef int _check_credentials(self) except -1
    cdef int _copy(self, ConnectParamsImpl other_params) except -1
    cdef bytes _get_new_password(self)
    cdef bytearray _get_obfuscator(self, str secret_value)
    cdef bytes _get_password(self)
    cdef str _get_private_key(self)
    cdef str _get_token(self)
    cdef object _get_token_expires(self, str token)
    cdef str _get_wallet_password(self)
    cdef int _parse_connect_string(self, str connect_string) except -1
    cdef int _set_access_token(self, object val, int error_num) except -1
    cdef int _set_access_token_param(self, object val) except -1
    cdef int _set_new_password(self, str password) except -1
    cdef int _set_password(self, str password) except -1
    cdef int _set_wallet_password(self, str password) except -1
    cdef bytearray _xor_bytes(self, bytearray a, bytearray b)


cdef class PoolParamsImpl(ConnectParamsImpl):
    cdef:
        public uint32_t min
        public uint32_t max
        public uint32_t increment
        public type connectiontype
        public uint32_t getmode
        public bint homogeneous
        public uint32_t timeout
        public uint32_t wait_timeout
        public uint32_t max_lifetime_session
        public object session_callback
        public uint32_t max_sessions_per_shard
        public bint soda_metadata_cache
        public int ping_interval
        public uint32_t ping_timeout


cdef class BaseConnImpl:
    cdef:
        readonly bint thin
        readonly str username
        readonly str dsn
        readonly str proxy_user
        public object inputtypehandler
        public object outputtypehandler
        public object warning
        public bint autocommit
        public bint invoke_session_callback
        readonly tuple server_version
        readonly bint supports_bool
        ssize_t _oson_max_fname_size
        bint _allow_bind_str_to_lob

    cdef object _check_value(self, DbType dbtype, BaseDbObjectTypeImpl objtype,
                             object value, bint* is_ok)
    cdef BaseCursorImpl _create_cursor_impl(self)


cdef class BasePoolImpl:
    cdef:
        readonly str dsn
        readonly bint homogeneous
        readonly uint32_t increment
        readonly uint32_t min
        readonly uint32_t max
        readonly str username
        readonly str name
        ConnectParamsImpl connect_params


cdef class BaseCursorImpl:
    cdef:
        readonly str statement
        readonly uint64_t rowcount
        public uint32_t arraysize
        public uint32_t prefetchrows
        public object inputtypehandler
        public object outputtypehandler
        public object rowfactory
        public bint scrollable
        public bint set_input_sizes
        public list fetch_info_impls
        public list fetch_vars
        public list fetch_var_impls
        public list bind_vars
        public type bind_style
        public dict bind_vars_by_name
        public object warning
        uint32_t _buffer_rowcount
        uint32_t _buffer_index
        uint32_t _fetch_array_size
        bint _more_rows_to_fetch

    cdef int _bind_values(self, object cursor, object type_handler,
                          object params, uint32_t num_rows, uint32_t row_num,
                          bint defer_type_assignment) except -1
    cdef int _bind_values_by_name(self, object cursor, object type_handler,
                                  dict params, uint32_t num_rows,
                                  uint32_t row_num,
                                  bint defer_type_assignment) except -1
    cdef int _bind_values_by_position(self, object cursor, object type_handler,
                                      object params, uint32_t num_rows,
                                      uint32_t row_num,
                                      bint defer_type_assignment) except -1
    cdef int _check_binds(self, uint32_t num_execs) except -1
    cdef int _close(self, bint in_del) except -1
    cdef BaseVarImpl _create_fetch_var(self, object conn, object cursor,
                                       object type_handler, bint
                                       uses_fetch_info, ssize_t pos,
                                       FetchInfoImpl fetch_info)
    cdef object _create_row(self)
    cdef BaseVarImpl _create_var_impl(self, object conn)
    cdef int _fetch_rows(self, object cursor) except -1
    cdef BaseConnImpl _get_conn_impl(self)
    cdef object _get_input_type_handler(self)
    cdef object _get_output_type_handler(self, bint* uses_fetch_info)
    cdef int _init_fetch_vars(self, uint32_t num_columns) except -1
    cdef bint _is_plsql(self)
    cdef int _perform_binds(self, object conn, uint32_t num_execs) except -1
    cdef int _prepare(self, str statement, str tag,
                      bint cache_statement) except -1
    cdef int _reset_bind_vars(self, uint32_t num_rows) except -1
    cdef int _verify_var(self, object var) except -1
    cdef int bind_many(self, object cursor, list parameters) except -1
    cdef int bind_one(self, object cursor, object parameters) except -1


cdef class FetchInfoImpl:
    cdef:
        readonly int16_t precision
        readonly int16_t scale
        readonly uint32_t buffer_size
        readonly uint32_t size
        readonly bint nulls_allowed
        readonly str name
        readonly DbType dbtype
        readonly BaseDbObjectTypeImpl objtype
        readonly bint is_json
        readonly bint is_oson
        readonly str domain_schema
        readonly str domain_name
        readonly dict annotations
        readonly uint32_t vector_dimensions
        readonly uint8_t vector_format
        readonly uint8_t vector_flags


cdef class BaseVarImpl:
    cdef:
        readonly str name
        readonly int16_t precision
        readonly int16_t scale
        readonly uint32_t num_elements
        readonly object inconverter
        readonly object outconverter
        readonly uint32_t size
        readonly uint32_t buffer_size
        readonly bint bypass_decode
        readonly str encoding_errors
        readonly bint is_array
        readonly bint nulls_allowed
        readonly bint convert_nulls
        public uint32_t num_elements_in_array
        readonly DbType dbtype
        readonly BaseDbObjectTypeImpl objtype
        BaseConnImpl _conn_impl
        int _preferred_num_type
        FetchInfoImpl _fetch_info
        list _values
        bint _is_value_set

    cdef int _bind(self, object conn, BaseCursorImpl cursor,
                   uint32_t num_execs, object name, uint32_t pos) except -1
    cdef int _check_and_set_scalar_value(self, uint32_t pos, object value,
                                         bint* was_set) except -1
    cdef int _check_and_set_value(self, uint32_t pos, object value,
                                  bint* was_set) except -1
    cdef int _finalize_init(self) except -1
    cdef list _get_array_value(self)
    cdef object _get_scalar_value(self, uint32_t pos)
    cdef int _on_reset_bind(self, uint32_t num_rows) except -1
    cdef int _resize(self, uint32_t new_size) except -1
    cdef int _set_scalar_value(self, uint32_t pos, object value) except -1
    cdef int _set_num_elements_in_array(self, uint32_t num_elements) except -1
    cdef int _set_type_info_from_type(self, object typ) except -1
    cdef int _set_type_info_from_value(self, object value,
                                       bint is_plsql) except -1


cdef class BaseLobImpl:
    cdef:
        readonly DbType dbtype


cdef class BaseDbObjectTypeImpl:
    cdef:
        readonly str schema
        readonly str name
        readonly str package_name
        readonly list attrs
        readonly bint is_collection
        readonly dict attrs_by_name
        readonly DbType element_dbtype
        readonly BaseDbObjectTypeImpl element_objtype
        readonly int8_t element_precision
        readonly int8_t element_scale
        readonly uint32_t element_max_size
        readonly BaseConnImpl _conn_impl
        int _element_preferred_num_type

    cpdef str _get_fqn(self)


cdef class BaseDbObjectAttrImpl:
    cdef:
        readonly str name
        readonly DbType dbtype
        readonly BaseDbObjectTypeImpl objtype
        readonly int8_t precision
        readonly int8_t scale
        readonly uint32_t max_size
        int _preferred_num_type


cdef class BaseDbObjectImpl:
    cdef:
        readonly BaseDbObjectTypeImpl type

    cdef int _check_max_size(self, object value, uint32_t max_size,
                             ssize_t* actual_size, bint* violated) except -1


cdef class BaseSodaDbImpl:
    cdef:
        public bint supports_json
        object _conn


cdef class BaseSodaCollImpl:
    cdef:
        readonly str name


cdef class BaseSodaDocImpl:
    pass


cdef class BaseSodaDocCursorImpl:
    pass


cdef class BaseQueueImpl:
    cdef:
        readonly str name
        readonly BaseDbObjectTypeImpl payload_type
        readonly BaseDeqOptionsImpl deq_options_impl
        readonly BaseEnqOptionsImpl enq_options_impl
        readonly bint is_json


cdef class BaseDeqOptionsImpl:
    pass


cdef class BaseEnqOptionsImpl:
    pass


cdef class BaseMsgPropsImpl:
    cdef:
        public object payload


cdef class BaseSubscrImpl:
    cdef:
        readonly object callback
        readonly object connection
        readonly uint32_t namespace
        readonly str name
        readonly uint32_t protocol
        readonly str ip_address
        readonly uint32_t port
        readonly uint32_t timeout
        readonly uint32_t operations
        readonly uint32_t qos
        readonly uint64_t id
        readonly uint8_t grouping_class
        readonly uint32_t grouping_value
        readonly uint8_t grouping_type
        readonly bint client_initiated


cdef class BindVar:
    cdef:
        object var
        BaseVarImpl var_impl
        object name
        ssize_t pos
        bint has_value

    cdef int _create_var_from_type(self, object conn,
                                   BaseCursorImpl cursor_impl,
                                   object value) except -1
    cdef int _create_var_from_value(self, object conn,
                                    BaseCursorImpl cursor_impl, object value,
                                    uint32_t num_elements) except -1
    cdef int _set_by_type(self, object conn, BaseCursorImpl cursor_impl,
                          object typ) except -1
    cdef int _set_by_value(self, object conn, BaseCursorImpl cursor_impl,
                           object cursor, object value, object type_handler,
                           uint32_t row_num, uint32_t num_elements,
                           bint defer_type_assignment) except -1


cdef class BaseParser:

    cdef:
        ssize_t pos, temp_pos, num_chars
        str data_as_str
        int data_kind
        void *data

    cdef Py_UCS4 get_current_char(self)
    cdef int initialize(self, str data_to_parse) except -1
    cdef int parse_keyword(self) except -1
    cdef int parse_quoted_string(self, Py_UCS4 quote_type) except -1
    cdef int skip_spaces(self) except -1


cdef class PipelineImpl:
    cdef:
        readonly list operations


cdef class PipelineOpImpl:
    cdef:
        readonly str statement
        readonly str name
        readonly object parameters
        readonly object keyword_parameters
        readonly object return_type
        readonly object rowfactory
        readonly uint32_t arraysize
        readonly uint32_t num_rows
        readonly uint8_t op_type
        uint32_t num_execs


cdef class PipelineOpResultImpl:
    cdef:
        readonly PipelineOpImpl operation
        readonly object return_value
        readonly list rows
        readonly object error

    cdef int _capture_err(self, Exception exc) except -1


cdef int get_preferred_num_type(int16_t precision, int8_t scale)
cdef void pack_uint16(char_type *buf, uint16_t x, int order)
cdef void pack_uint32(char_type *buf, uint32_t x, int order)
cdef void pack_uint64(char_type *buf, uint64_t x, int order)
cdef uint16_t unpack_uint16(const char_type *buf, int order)
cdef uint32_t unpack_uint32(const char_type *buf, int order)



# -----------------------------------------------------------------------------
# Copyright (c) 2024, Oracle and/or its affiliates.
#
# This software is dual-licensed to you under the Universal Permissive License
# (UPL) 1.0 as shown at https://oss.oracle.com/licenses/upl and Apache License
# 2.0 as shown at http://www.apache.org/licenses/LICENSE-2.0. You may choose
# either license.
#
# If you elect to accept the software under the Apache License, Version 2.0,
# the following applies:
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# -----------------------------------------------------------------------------

"""
7100 - Module for testing interval year to month variables
"""

import oracledb
import test_env


class TestCase(test_env.BaseTestCase):
    def setUp(self):
        super().setUp()
        self.raw_data = []
        self.data_by_key = {}
        for i in range(1, 11):
            delta = oracledb.IntervalYM(i - 5, -i if i - 5 < 0 else i)
            if i % 2 == 0:
                nullable_delta = None
            else:
                nullable_delta = oracledb.IntervalYM(i + 5, i + 2)
            precision_col = oracledb.IntervalYM(3, 8)
            data_tuple = (i, delta, nullable_delta, precision_col)
            self.raw_data.append(data_tuple)
            self.data_by_key[i] = data_tuple

    def test_7100(self):
        "7100 - test binding in an interval"
        value = oracledb.IntervalYM(1, 6)
        self.cursor.execute(
            "select * from TestIntervalYMs where IntervalCol = :value",
            value=value,
        )
        self.assertEqual(self.cursor.fetchall(), [self.data_by_key[6]])

    def test_7101(self):
        "7101 - test binding in a null"
        self.cursor.setinputsizes(value=oracledb.DB_TYPE_INTERVAL_YM)
        self.cursor.execute(
            "select * from TestIntervalYMs where IntervalCol = :value",
            value=None,
        )
        self.assertEqual(self.cursor.fetchall(), [])

    def test_7102(self):
        "7102 - test binding out with set input sizes defined"
        bind_vars = self.cursor.setinputsizes(
            value=oracledb.DB_TYPE_INTERVAL_YM
        )
        self.cursor.execute(
            """
            begin
                :value := to_yminterval('-25-7');
            end;
            """
        )
        expected_value = oracledb.IntervalYM(years=-25, months=-7)
        self.assertEqual(bind_vars["value"].getvalue(), expected_value)

    def test_7103(self):
        "7103 - test binding in/out with set input sizes defined"
        bind_vars = self.cursor.setinputsizes(
            value=oracledb.DB_TYPE_INTERVAL_YM
        )
        self.cursor.execute(
            """
            begin
                :value := :value + to_yminterval('3-8');
            end;
            """,
            value=oracledb.IntervalYM(years=8, months=4),
        )
        expected_value = oracledb.IntervalYM(years=12, months=0)
        self.assertEqual(bind_vars["value"].getvalue(), expected_value)

    def test_7104(self):
        "7104 - test binding out with cursor.var() method"
        var = self.cursor.var(oracledb.DB_TYPE_INTERVAL_YM)
        self.cursor.execute(
            """
            begin
                :value := to_yminterval('1-9');
            end;
            """,
            value=var,
        )
        expected_value = oracledb.IntervalYM(years=1, months=9)
        self.assertEqual(var.getvalue(), expected_value)

    def test_7105(self):
        "7105 - test binding in/out with cursor.var() method"
        var = self.cursor.var(oracledb.DB_TYPE_INTERVAL_YM)
        var.setvalue(0, oracledb.IntervalYM(years=3, months=10))
        self.cursor.execute(
            """
            begin
                :value := :value + to_yminterval('2-5');
            end;
            """,
            value=var,
        )
        expected_value = oracledb.IntervalYM(years=6, months=3)
        self.assertEqual(var.getvalue(), expected_value)

    def test_7106(self):
        "7106 - test cursor description is accurate"
        self.cursor.execute("select * from TestIntervalYMs")
        expected_value = [
            ("INTCOL", oracledb.DB_TYPE_NUMBER, 10, None, 9, 0, False),
            (
                "INTERVALCOL",
                oracledb.DB_TYPE_INTERVAL_YM,
                None,
                None,
                2,
                0,
                False,
            ),
            (
                "NULLABLECOL",
                oracledb.DB_TYPE_INTERVAL_YM,
                None,
                None,
                2,
                0,
                True,
            ),
            (
                "INTERVALPRECISIONCOL",
                oracledb.DB_TYPE_INTERVAL_YM,
                None,
                None,
                3,
                0,
                True,
            ),
        ]
        self.assertEqual(self.cursor.description, expected_value)

    def test_7107(self):
        "7107 - test that fetching all of the data returns the correct results"
        self.cursor.execute("select * From TestIntervalYMs order by IntCol")
        self.assertEqual(self.cursor.fetchall(), self.raw_data)
        self.assertEqual(self.cursor.fetchall(), [])

    def test_7108(self):
        "7108 - test that fetching data in chunks returns the correct results"
        self.cursor.execute("select * From TestIntervalYMs order by IntCol")
        self.assertEqual(self.cursor.fetchmany(3), self.raw_data[0:3])
        self.assertEqual(self.cursor.fetchmany(2), self.raw_data[3:5])
        self.assertEqual(self.cursor.fetchmany(4), self.raw_data[5:9])
        self.assertEqual(self.cursor.fetchmany(3), self.raw_data[9:])
        self.assertEqual(self.cursor.fetchmany(3), [])

    def test_7109(self):
        "7109 - test that fetching a single row returns the correct results"
        self.cursor.execute(
            """
            select *
            from TestIntervalYMs
            where IntCol in (3, 4)
            order by IntCol
            """
        )
        self.assertEqual(self.cursor.fetchone(), self.data_by_key[3])
        self.assertEqual(self.cursor.fetchone(), self.data_by_key[4])
        self.assertIsNone(self.cursor.fetchone())

    def test_7110(self):
        "7110 - test binding and fetching a negative interval"
        value = oracledb.IntervalYM(years=-12, months=-5)
        self.cursor.execute("select :1 from dual", [value])
        (result,) = self.cursor.fetchone()
        self.assertEqual(result, value)


if __name__ == "__main__":
    test_env.run_test_cases()



#------------------------------------------------------------------------------
# Copyright (c) 2020, 2024, Oracle and/or its affiliates.
#
# This software is dual-licensed to you under the Universal Permissive License
# (UPL) 1.0 as shown at https://oss.oracle.com/licenses/upl and Apache License
# 2.0 as shown at http://www.apache.org/licenses/LICENSE-2.0. You may choose
# either license.
#
# If you elect to accept the software under the Apache License, Version 2.0,
# the following applies:
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#------------------------------------------------------------------------------

#------------------------------------------------------------------------------
# thick_impl.pyx
#
# Cython file for interfacing with ODPI-C.
#------------------------------------------------------------------------------

# cython: language_level=3

cimport cython
cimport cpython
cimport cpython.datetime as cydatetime
from cpython cimport array

import array
import datetime
import decimal

cydatetime.import_datetime()

from .base_impl cimport (
    BaseConnImpl,
    BaseCursorImpl,
    BaseDbObjectImpl,
    BaseDbObjectAttrImpl,
    BaseDbObjectTypeImpl,
    BaseDeqOptionsImpl,
    BaseEnqOptionsImpl,
    BaseLobImpl,
    BaseMsgPropsImpl,
    BasePoolImpl,
    BaseQueueImpl,
    BaseSodaCollImpl,
    BaseSodaDbImpl,
    BaseSodaDocImpl,
    BaseSodaDocCursorImpl,
    BaseSubscrImpl,
    BaseVarImpl,
    BindVar,
    C_DEFAULTS,
    ConnectParamsImpl,
    DbType,
    DB_TYPE_NUM_CURSOR,
    DRIVER_NAME,
    DRIVER_VERSION,
    DRIVER_INSTALLATION_URL,
    ENCODING_UTF8,
    PURITY_DEFAULT,
    PY_TYPE_DATE,
    PY_TYPE_DATETIME,
    PY_TYPE_DB_OBJECT,
    PY_TYPE_DECIMAL,
    PY_TYPE_JSON_ID,
    PY_TYPE_INTERVAL_YM,
    PY_TYPE_LOB,
    PY_TYPE_MESSAGE,
    PY_TYPE_MESSAGE_QUERY,
    PY_TYPE_MESSAGE_ROW,
    PY_TYPE_MESSAGE_TABLE,
    PY_TYPE_TIMEDELTA,
    FetchInfoImpl,
    PoolParamsImpl,
    get_preferred_num_type,
    NUM_TYPE_FLOAT,
    NUM_TYPE_INT,
    NUM_TYPE_DECIMAL,
    VectorDecoder,
    VectorEncoder,
)
from libc.string cimport memchr, memcpy, memset

include "impl/thick/odpi.pxd"

cdef struct DriverInfo:
    dpiContext *context
    dpiVersionInfo client_version_info
    bint soda_use_json_desc

cdef DriverInfo driver_info = \
        DriverInfo(NULL, dpiVersionInfo(0, 0, 0, 0, 0, 0), True)

driver_context_params = None

include "impl/thick/buffer.pyx"
include "impl/thick/connection.pyx"
include "impl/thick/pool.pyx"
include "impl/thick/cursor.pyx"
include "impl/thick/lob.pyx"
include "impl/thick/json.pyx"
include "impl/thick/var.pyx"
include "impl/thick/dbobject.pyx"
include "impl/thick/soda.pyx"
include "impl/thick/queue.pyx"
include "impl/thick/subscr.pyx"
include "impl/thick/utils.pyx"



#------------------------------------------------------------------------------
# Copyright (c) 2020, 2024, Oracle and/or its affiliates.
#
# This software is dual-licensed to you under the Universal Permissive License
# (UPL) 1.0 as shown at https://oss.oracle.com/licenses/upl and Apache License
# 2.0 as shown at http://www.apache.org/licenses/LICENSE-2.0. You may choose
# either license.
#
# If you elect to accept the software under the Apache License, Version 2.0,
# the following applies:
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#------------------------------------------------------------------------------

#------------------------------------------------------------------------------
# dbobject.pyx
#
# Cython file defining the base DbObjectType, DbObjectAttr and DbObject
# implementation classes (embedded in base_impl.pyx).
#------------------------------------------------------------------------------

cdef class BaseDbObjectImpl:

    cdef int _check_max_size(self, object value, uint32_t max_size,
                             ssize_t* actual_size, bint* violated) except -1:
        """
        Checks to see if the maximum size has been violated.
        """
        violated[0] = False
        if max_size > 0 and value is not None:
            if isinstance(value, str):
                actual_size[0] = len((<str> value).encode())
            else:
                actual_size[0] = len(<bytes> value)
            if actual_size[0] > max_size:
                violated[0] = True

    def append(self, object value):
        """
        Appends a value to the collection after first checking to see if the
        value is acceptable.
        """
        cdef:
            BaseConnImpl conn_impl = self.type._conn_impl
            ssize_t actual_size
            bint violated
        value = conn_impl._check_value(self.type.element_dbtype,
                                       self.type.element_objtype, value, NULL)
        self._check_max_size(value, self.type.element_max_size, &actual_size,
                             &violated)
        if violated:
            errors._raise_err(errors.ERR_DBOBJECT_ELEMENT_MAX_SIZE_VIOLATED,
                              index=self.get_size(),
                              type_name=self.type._get_fqn(),
                              actual_size=actual_size,
                              max_size=self.type.element_max_size)
        self.append_checked(value)

    def append_checked(self, object value):
        errors._raise_not_supported("appending a value to a collection")

    def copy(self):
        errors._raise_not_supported("creating a copy of an object")

    def delete_by_index(self, int32_t index):
        errors._raise_not_supported("deleting an element in a collection")

    def exists_by_index(self, int32_t index):
        errors._raise_not_supported(
            "determining if an entry exists in a collection"
        )

    def get_attr_value(self, BaseDbObjectAttrImpl attr):
        errors._raise_not_supported("getting an attribute value")

    def get_element_by_index(self, int32_t index):
        errors._raise_not_supported("getting an element of a collection")

    def get_first_index(self):
        errors._raise_not_supported("getting the first index of a collection")

    def get_last_index(self):
        errors._raise_not_supported("getting the last index of a collection")

    def get_next_index(self, int32_t index):
        errors._raise_not_supported("getting the next index of a collection")

    def get_prev_index(self, int32_t index):
        errors._raise_not_supported(
            "getting the previous index of a collection"
        )

    def get_size(self):
        errors._raise_not_supported("getting the size of a collection")

    def set_attr_value(self, BaseDbObjectAttrImpl attr, object value):
        """
        Sets the attribute value after first checking to see if the value is
        acceptable.
        """
        cdef:
            BaseConnImpl conn_impl = self.type._conn_impl
            ssize_t actual_size
            bint violated
        value = conn_impl._check_value(attr.dbtype, attr.objtype, value, NULL)
        self._check_max_size(value, attr.max_size, &actual_size, &violated)
        if violated:
            errors._raise_err(errors.ERR_DBOBJECT_ATTR_MAX_SIZE_VIOLATED,
                              attr_name=attr.name,
                              type_name=self.type._get_fqn(),
                              actual_size=actual_size, max_size=attr.max_size)
        self.set_attr_value_checked(attr, value)

    def set_attr_value_checked(self, BaseDbObjectAttrImpl attr, object value):
        errors._raise_not_supported("setting an attribute value")

    def set_element_by_index(self, int32_t index, object value):
        """
        Sets the element value after first checking to see if the value is
        acceptable.
        """
        cdef:
            BaseConnImpl conn_impl = self.type._conn_impl
            ssize_t actual_size
            bint violated
        value = conn_impl._check_value(self.type.element_dbtype,
                                       self.type.element_objtype, value, NULL)
        self._check_max_size(value, self.type.element_max_size, &actual_size,
                             &violated)
        if violated:
            errors._raise_err(errors.ERR_DBOBJECT_ELEMENT_MAX_SIZE_VIOLATED,
                              index=index, type_name=self.type._get_fqn(),
                              actual_size=actual_size,
                              max_size=self.type.element_max_size)
        self.set_element_by_index_checked(index, value)

    def set_element_by_index_checked(self, int32_t index, object value):
        errors._raise_not_supported("setting an element of a collection")

    def trim(self, int32_t num_to_trim):
        errors._raise_not_supported("trimming elements from a collection")


cdef class BaseDbObjectAttrImpl:
    pass


cdef class BaseDbObjectTypeImpl:

    def __eq__(self, other):
        if isinstance(other, BaseDbObjectTypeImpl):
            return other._conn_impl is self._conn_impl \
                    and other.schema == self.schema \
                    and other.name == self.name
        return NotImplemented

    cpdef str _get_fqn(self):
        """
        Return the fully qualified name of the type.
        """
        if self.package_name is not None:
            return f"{self.schema}.{self.package_name}.{self.name}"
        return f"{self.schema}.{self.name}"

    def create_new_object(self):
        errors._raise_not_supported("creating a new object")



# -----------------------------------------------------------------------------
# Copyright (c) 2024, Oracle and/or its affiliates.
#
# This software is dual-licensed to you under the Universal Permissive License
# (UPL) 1.0 as shown at https://oss.oracle.com/licenses/upl and Apache License
# 2.0 as shown at http://www.apache.org/licenses/LICENSE-2.0. You may choose
# either license.
#
# If you elect to accept the software under the Apache License, Version 2.0,
# the following applies:
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# -----------------------------------------------------------------------------

"""
E1600 - Module for testing the generation of warnings. No special setup is
required but sys privilege is required in order to create and drop users.
"""

import time

import oracledb
import test_env


class TestCase(test_env.BaseTestCase):
    profile_name = "profile_ext_test_1600"
    user_name = "user_ext_test_1600"
    requires_connection = False

    @classmethod
    def setUpClass(cls):
        conn = test_env.get_admin_connection()
        password = test_env.get_main_password()
        cursor = conn.cursor()
        cursor.execute(
            f"""
            declare
                e_user_missing exception;
                pragma exception_init(e_user_missing, -1918);
            begin
                execute immediate('drop user {cls.user_name} cascade');
            exception
            when e_user_missing then
                null;
            end;
            """
        )
        cursor.execute(
            f"""
            declare
                e_user_missing exception;
                pragma exception_init(e_user_missing, -2380);
            begin
                execute immediate('drop profile {cls.profile_name}');
            exception
            when e_user_missing then
                null;
            end;
            """
        )
        cursor.execute(f"create user {cls.user_name} identified by {password}")
        cursor.execute(f"grant create session to {cls.user_name}")
        cursor.execute(
            f"""
            create profile {cls.profile_name} limit
            password_life_time 1 / 24 / 60 / 60
            password_grace_time 7
            """
        )
        cursor.execute(
            f"alter user {cls.user_name} profile {cls.profile_name}"
        )
        time.sleep(2)

    @classmethod
    def tearDownClass(cls):
        conn = test_env.get_admin_connection()
        cursor = conn.cursor()
        cursor.execute(f"drop user {cls.user_name} cascade")
        cursor.execute(f"drop profile {cls.profile_name}")

    def test_ext_1600(self):
        "E1600 - test standalone connection generates a warning"
        password = test_env.get_main_password()
        with oracledb.connect(
            user=self.user_name,
            password=password,
            dsn=test_env.get_connect_string(),
        ) as conn:
            self.assertIn(conn.warning.full_code, ["ORA-28002", "ORA-28098"])

    def test_ext_1601_pooled_conn_warning_min_0(self):
        "E1601 - test pooled connection generates a warning (min 0)"
        password = test_env.get_main_password()
        pool = oracledb.create_pool(
            user=self.user_name,
            password=password,
            dsn=test_env.get_connect_string(),
            min=0,
            max=5,
            increment=1,
        )
        with pool.acquire() as conn:
            self.assertIn(conn.warning.full_code, ["ORA-28002", "ORA-28098"])
        with pool.acquire() as conn:
            self.assertIsNone(conn.warning)
        pool.close(0)

    def test_ext_1602_pooled_conn_warning_min_1(self):
        "E1602 - test pooled connection generates a warning (min 1)"
        password = test_env.get_main_password()
        pool = oracledb.create_pool(
            user=self.user_name,
            password=password,
            dsn=test_env.get_connect_string(),
            min=1,
            max=5,
            increment=1,
        )
        with pool.acquire() as conn:
            self.assertIn(conn.warning.full_code, ["ORA-28002", "ORA-28098"])
        with pool.acquire() as conn:
            self.assertIsNone(conn.warning)
        pool.close(0)


if __name__ == "__main__":
    test_env.run_test_cases()



# -----------------------------------------------------------------------------
# Copyright (c) 2024, Oracle and/or its affiliates.
#
# This software is dual-licensed to you under the Universal Permissive License
# (UPL) 1.0 as shown at https://oss.oracle.com/licenses/upl and Apache License
# 2.0 as shown at http://www.apache.org/licenses/LICENSE-2.0. You may choose
# either license.
#
# If you elect to accept the software under the Apache License, Version 2.0,
# the following applies:
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# -----------------------------------------------------------------------------

"""
E1400 - Module for testing inband notification. No special setup is required
but the test makes use of debugging packages that are not intended for normal
use.
"""

import test_env


class TestCase(test_env.BaseTestCase):
    requires_connection = False

    @classmethod
    def setUpClass(cls):
        cls.admin_conn = test_env.get_admin_connection()
        user = test_env.get_main_user()
        with cls.admin_conn.cursor() as cursor:
            cursor.execute(f"grant execute on dbms_tg_dbg to {user}")

    @classmethod
    def tearDownClass(cls):
        user = test_env.get_main_user()
        with cls.admin_conn.cursor() as cursor:
            cursor.execute(f"revoke execute on dbms_tg_dbg from {user}")

    def test_ext_1400(self):
        "E1400 - test standalone connection is marked unhealthy"
        conn = test_env.get_connection()
        self.assertEqual(conn.is_healthy(), True)
        with conn.cursor() as cursor:
            cursor.callproc("dbms_tg_dbg.set_session_drainable")
            cursor.execute("select user from dual")
            (user,) = cursor.fetchone()
            self.assertEqual(user, test_env.get_main_user().upper())
        self.assertEqual(conn.is_healthy(), False)

    def test_ext_1401(self):
        "E1401 - test pooled connection that is marked unhealthy"
        pool = test_env.get_pool(min=1, max=1, increment=1)
        with pool.acquire() as conn:
            self.assertEqual(conn.is_healthy(), True)
            with conn.cursor() as cursor:
                cursor.callproc("dbms_tg_dbg.set_session_drainable")
                info = self.get_sid_serial(conn)
            self.assertEqual(conn.is_healthy(), False)
        with pool.acquire() as conn:
            self.assertEqual(conn.is_healthy(), True)
            new_info = self.get_sid_serial(conn)
            self.assertNotEqual(new_info, info)


if __name__ == "__main__":
    test_env.run_test_cases()



# -----------------------------------------------------------------------------
# Copyright (c) 2024, Oracle and/or its affiliates.
#
# This software is dual-licensed to you under the Universal Permissive License
# (UPL) 1.0 as shown at https://oss.oracle.com/licenses/upl and Apache License
# 2.0 as shown at http://www.apache.org/licenses/LICENSE-2.0. You may choose
# either license.
#
# If you elect to accept the software under the Apache License, Version 2.0,
# the following applies:
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# -----------------------------------------------------------------------------

"""
7400 - Module for testing TPC (two-phase commit) transactions with asyncio.
"""

import unittest

import oracledb
import test_env


@unittest.skipUnless(
    test_env.get_is_thin(), "asyncio not supported in thick mode"
)
class TestCase(test_env.BaseAsyncTestCase):
    async def test_7400(self):
        "7400 - test begin, prepare, roll back global transaction"
        await self.cursor.execute("truncate table TestTempTable")
        xid = self.conn.xid(3900, b"txn3900", b"branchId")
        await self.conn.tpc_begin(xid)
        self.assertEqual(await self.conn.tpc_prepare(), False)
        await self.conn.tpc_begin(xid)
        await self.cursor.execute(
            """
            insert into TestTempTable (IntCol, StringCol1)
            values (1, 'tesName')
            """
        )
        self.assertEqual(await self.conn.tpc_prepare(), True)
        await self.conn.tpc_rollback()
        await self.cursor.execute("select count(*) from TestTempTable")
        (count,) = await self.cursor.fetchone()
        self.assertEqual(count, 0)

    async def test_7401(self):
        "7401 - test begin, prepare, commit global transaction"
        await self.cursor.execute("truncate table TestTempTable")
        xid = self.conn.xid(3901, "txn3901", "branchId")
        await self.conn.tpc_begin(xid)
        await self.cursor.execute(
            """
            insert into TestTempTable (IntCol, StringCol1)
            values (1, 'tesName')
            """
        )
        self.assertEqual(await self.conn.tpc_prepare(), True)
        await self.conn.tpc_commit()
        await self.cursor.execute(
            "select IntCol, StringCol1 from TestTempTable"
        )
        self.assertEqual(await self.cursor.fetchall(), [(1, "tesName")])

    async def test_7402(self):
        "7402 - test multiple global transactions on the same connection"
        await self.cursor.execute("truncate table TestTempTable")
        xid1 = self.conn.xid(3902, "txn3902", "branch1")
        xid2 = self.conn.xid(3902, b"txn3902", b"branch2")
        await self.conn.tpc_begin(xid1)
        await self.cursor.execute(
            """
            insert into TestTempTable (IntCol, StringCol1)
            values (1, 'tesName')
            """
        )
        await self.conn.tpc_end()
        await self.conn.tpc_begin(xid2)
        await self.cursor.execute(
            """
            insert into TestTempTable (IntCol, StringCol1)
            values (2, 'tesName')
            """
        )
        await self.conn.tpc_end()
        needs_commit1 = await self.conn.tpc_prepare(xid1)
        needs_commit2 = await self.conn.tpc_prepare(xid2)
        if needs_commit1:
            await self.conn.tpc_commit(xid1)
        if needs_commit2:
            await self.conn.tpc_commit(xid2)
        await self.cursor.execute(
            "select IntCol, StringCol1 from TestTempTable order by IntCol"
        )
        expected_rows = [(1, "tesName"), (2, "tesName")]
        self.assertEqual(await self.cursor.fetchall(), expected_rows)

    async def test_7403(self):
        "7403 - test rollback with parameter xid"
        await self.cursor.execute("truncate table TestTempTable")
        xid1 = self.conn.xid(3901, b"txn3901", b"branch1")
        xid2 = self.conn.xid(3902, "txn3902", "branch2")
        for count, xid in enumerate([xid1, xid2]):
            await self.conn.tpc_begin(xid)
            await self.cursor.execute(
                """
                insert into TestTempTable (IntCol, StringCol1)
                values (:id, 'tesName')
                """,
                id=count,
            )
            await self.conn.tpc_end()
        await self.conn.tpc_rollback(xid1)

        with self.assertRaisesFullCode("ORA-24756"):
            await self.conn.tpc_prepare(xid1)
        needs_commit = await self.conn.tpc_prepare(xid2)
        if needs_commit:
            await self.conn.tpc_commit(xid2)
        await self.cursor.execute(
            "select IntCol, StringCol1 from TestTempTable order by IntCol"
        )
        self.assertEqual(await self.cursor.fetchall(), [(1, "tesName")])

    async def test_7404(self):
        "7404 - test resuming a transaction"
        await self.cursor.execute("truncate table TestTempTable")
        xid1 = self.conn.xid(3939, "txn3939", "branch39")
        xid2 = self.conn.xid(3940, "txn3940", "branch40")
        values = [[xid1, (1, "User Info")], [xid2, (2, "Other User Info")]]
        for xid, data in values:
            await self.conn.tpc_begin(xid)
            await self.cursor.execute(
                """
                insert into TestTempTable (IntCol, StringCol1)
                values (:1, :2)
                """,
                data,
            )
            await self.conn.tpc_end()
        for xid, data in values:
            await self.conn.tpc_begin(xid, oracledb.TPC_BEGIN_RESUME)
            await self.cursor.execute(
                "select IntCol, StringCol1 from TestTempTable"
            )
            (res,) = await self.cursor.fetchall()
            self.assertEqual(res, data)
            await self.conn.tpc_rollback(xid)

    async def test_7405(self):
        "7405 - test promoting a local transaction to a tpc transaction"
        await self.cursor.execute("truncate table TestTempTable")
        xid = self.conn.xid(3941, "txn3941", "branch41")
        values = (1, "String 1")
        await self.cursor.execute(
            "insert into TestTempTable (IntCol, StringCol1) values (:1, :2)",
            values,
        )
        with self.assertRaisesFullCode("ORA-24776"):
            await self.conn.tpc_begin(xid)
        await self.conn.tpc_begin(xid, oracledb.TPC_BEGIN_PROMOTE)
        await self.cursor.execute(
            "select IntCol, StringCol1 from TestTempTable"
        )
        (res,) = await self.cursor.fetchall()
        self.assertEqual(res, values)
        await self.conn.tpc_rollback(xid)

    async def test_7406(self):
        "7406 - test ending a transaction with parameter xid"
        await self.cursor.execute("truncate table TestTempTable")
        xid1 = self.conn.xid(7406, "txn7406a", "branch3")
        xid2 = self.conn.xid(7406, b"txn7406b", b"branch4")
        await self.conn.tpc_begin(xid1)
        await self.cursor.execute(
            """
            insert into TestTempTable (IntCol, StringCol1)
            values (1, 'test7406a')
            """
        )
        await self.conn.tpc_begin(xid2)
        with self.assertRaisesFullCode("ORA-24758"):
            await self.conn.tpc_end(xid1)
        await self.cursor.execute(
            """
            insert into TestTempTable (IntCol, StringCol1)
            values (2, 'test7406b')
            """
        )
        await self.conn.tpc_end(xid2)
        with self.assertRaisesFullCode("ORA-25352"):
            await self.conn.tpc_end(xid1)
        await self.conn.tpc_rollback(xid1)
        await self.conn.tpc_rollback(xid2)

    async def test_7407(self):
        "7407 - test tpc_recover()"
        await self.cursor.execute("truncate table TestTempTable")
        n_xids = 10
        for i in range(n_xids):
            xid = self.conn.xid(7407 + i, f"txn7407{i}", f"branch{i}")
            await self.conn.tpc_begin(xid)
            await self.cursor.execute(
                """
                insert into TestTempTable (IntCol, StringCol1)
                values (:1, 'test7407')
                """,
                [i + 1],
            )
            await self.conn.tpc_prepare(xid)
        recovers = await self.conn.tpc_recover()
        self.assertEqual(len(recovers), n_xids)

        await self.cursor.execute("select * from DBA_PENDING_TRANSACTIONS")
        self.assertEqual(await self.cursor.fetchall(), recovers)

        for xid in recovers:
            if xid.format_id % 2 == 0:
                await self.conn.tpc_commit(xid)
        recovers = await self.conn.tpc_recover()
        self.assertEqual(len(recovers), n_xids // 2)

        for xid in recovers:
            await self.conn.tpc_rollback(xid)
        recovers = await self.conn.tpc_recover()
        self.assertEqual(len(recovers), 0)

    async def test_7408(self):
        "7408 - test tpc_recover() with read-only transaction"
        await self.cursor.execute("truncate table TestTempTable")
        for i in range(4):
            xid = self.conn.xid(7408 + i, f"txn7408{i}", f"branch{i}")
            await self.conn.tpc_begin(xid)
            await self.cursor.execute("select * from TestTempTable")
            await self.conn.tpc_prepare(xid)
        recovers = await self.conn.tpc_recover()
        self.assertEqual(len(recovers), 0)

    async def test_7409(self):
        "7409 - test tpc_commit() with one_phase parameter"
        await self.cursor.execute("truncate table TestTempTable")
        xid = self.conn.xid(7409, "txn7409", "branch1")
        await self.conn.tpc_begin(xid)
        values = (1, "test7409")
        await self.cursor.execute(
            """
            insert into TestTempTable (IntCol, StringCol1)
            values (:1, :2)
            """,
            values,
        )
        await self.cursor.execute(
            "select IntCol, StringCol1 from TestTempTable"
        )
        await self.conn.tpc_commit(xid, one_phase=True)
        self.assertEqual(await self.cursor.fetchall(), [values])

    async def test_7410(self):
        "7410 - test negative cases for tpc_commit()"
        await self.cursor.execute("truncate table TestTempTable")
        xid = self.conn.xid(7410, "txn7410", "branch1")
        await self.conn.tpc_begin(xid)
        await self.cursor.execute(
            """
            insert into TestTempTable (IntCol, StringCol1)
            values (1, 'test7410')
            """
        )
        with self.assertRaises(TypeError):
            await self.conn.tpc_commit("invalid xid")
        await self.conn.tpc_prepare(xid)
        with self.assertRaisesFullCode("ORA-02053"):
            await self.conn.tpc_commit(xid, one_phase=True)
        with self.assertRaisesFullCode("ORA-24756"):
            await self.conn.tpc_commit(xid)

    async def test_7411(self):
        "7411 - test starting an already created transaction"
        await self.cursor.execute("truncate table TestTempTable")
        xid = self.conn.xid(7411, "txn7411", "branch1")
        await self.conn.tpc_begin(xid)
        with self.assertRaisesFullCode("ORA-24757"):
            await self.conn.tpc_begin(xid, oracledb.TPC_BEGIN_NEW)
        with self.assertRaisesFullCode("ORA-24797"):
            await self.conn.tpc_begin(xid, oracledb.TPC_BEGIN_PROMOTE)
        await self.conn.tpc_end()
        for flag in [oracledb.TPC_BEGIN_NEW, oracledb.TPC_BEGIN_PROMOTE]:
            with self.assertRaisesFullCode("ORA-24757"):
                await self.conn.tpc_begin(xid, flag)
        await self.conn.tpc_rollback(xid)

    async def test_7412(self):
        "7412 - test resuming a prepared transaction"
        await self.cursor.execute("truncate table TestTempTable")
        xid = self.conn.xid(7412, "txn7412", "branch1")
        await self.conn.tpc_begin(xid)
        await self.conn.tpc_prepare(xid)
        with self.assertRaisesFullCode("ORA-24756"):
            await self.conn.tpc_begin(xid, oracledb.TPC_BEGIN_RESUME)

    async def test_7413(self):
        "7413 - test tpc_begin and tpc_end with invalid parameters"
        await self.cursor.execute("truncate table TestTempTable")
        xid = self.conn.xid(7413, "txn7413", "branch1")
        test_values = [
            (self.conn.tpc_begin, "DPY-2050"),
            (self.conn.tpc_end, "DPY-2051"),
        ]
        for tpc_function, error_code in test_values:
            with self.assertRaises(TypeError):
                await tpc_function("invalid xid")
            with self.assertRaisesFullCode(error_code):
                await tpc_function(xid, "invalid flag")
            with self.assertRaisesFullCode(error_code):
                await tpc_function(xid, 70)

    async def test_7414(self):
        "7414 - test commiting transaction without tpc_commit"
        xid = self.conn.xid(7414, "txn7409", "branch1")
        await self.conn.tpc_begin(xid)
        with self.assertRaisesFullCode("ORA-02089"):
            await self.cursor.execute("truncate table TestTempTable")

    async def test_7415(self):
        "7415 - test tpc_commit when a commit is not needed"
        xid = self.conn.xid(7416, "txn7416", "branch1")
        await self.conn.tpc_begin(xid)
        await self.cursor.execute("select * from TestTempTable")
        await self.conn.tpc_end(xid)
        await self.conn.tpc_prepare(xid)
        with self.assertRaisesFullCode("ORA-24756"):
            await self.conn.tpc_commit(xid)

    async def test_7416(self):
        "7416 - test transaction_in_progress"
        await self.cursor.execute("truncate table TestTempTable")
        xid = self.conn.xid(7415, "txn7415", "branch1")
        self.assertFalse(self.conn.transaction_in_progress)

        await self.conn.tpc_begin(xid)
        self.assertTrue(self.conn.transaction_in_progress)
        await self.cursor.execute(
            "insert into TestTempTable (IntCol) values (2)"
        )

        await self.conn.tpc_end(xid)
        self.assertFalse(self.conn.transaction_in_progress)

        await self.conn.tpc_prepare(xid)
        self.assertFalse(self.conn.transaction_in_progress)

        await self.conn.tpc_commit(xid)
        self.assertFalse(self.conn.transaction_in_progress)


if __name__ == "__main__":
    test_env.run_test_cases()



#------------------------------------------------------------------------------
# Copyright (c) 2024, Oracle and/or its affiliates.
#
# This software is dual-licensed to you under the Universal Permissive License
# (UPL) 1.0 as shown at https://oss.oracle.com/licenses/upl and Apache License
# 2.0 as shown at http://www.apache.org/licenses/LICENSE-2.0. You may choose
# either license.
#
# If you elect to accept the software under the Apache License, Version 2.0,
# the following applies:
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#------------------------------------------------------------------------------

#------------------------------------------------------------------------------
# defaults.pyx
#
# Cython file defining the DefaultsImpl class (embedded in base_impl.pyx).
#------------------------------------------------------------------------------

cdef class DefaultsImpl:

    def __init__(self):
        self.arraysize = 100
        self.config_dir = os.environ.get("TNS_ADMIN")
        self.fetch_lobs = True
        self.fetch_decimals = False
        self.prefetchrows = 2
        self.stmtcachesize = 20

cdef DefaultsImpl C_DEFAULTS = DefaultsImpl()
DEFAULTS = C_DEFAULTS



#------------------------------------------------------------------------------
# Copyright (c) 2020, 2024, Oracle and/or its affiliates.
#
# This software is dual-licensed to you under the Universal Permissive License
# (UPL) 1.0 as shown at https://oss.oracle.com/licenses/upl and Apache License
# 2.0 as shown at http://www.apache.org/licenses/LICENSE-2.0. You may choose
# either license.
#
# If you elect to accept the software under the Apache License, Version 2.0,
# the following applies:
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#------------------------------------------------------------------------------

#------------------------------------------------------------------------------
# base_impl.pyx
#
# Cython file for the base implementation that the thin and thick
# implementations use.
#------------------------------------------------------------------------------

# cython: language_level=3

cimport cython
cimport cpython
cimport cpython.datetime as cydatetime

from libc.stdint cimport int8_t, int16_t, int32_t, int64_t
from libc.stdint cimport uint8_t, uint16_t, uint32_t, uint64_t
from libc.stdint cimport UINT8_MAX, UINT16_MAX, UINT32_MAX, UINT64_MAX
from libc.string cimport memcpy
from cpython cimport array

import array

import base64
import datetime
import decimal
import inspect
import json
import os
import random
import secrets
import ssl
import sys

cydatetime.import_datetime()

include "impl/base/types.pyx"

# Python types used by the driver
cdef type PY_TYPE_ASYNC_CURSOR
cdef type PY_TYPE_ASYNC_LOB
cdef type PY_TYPE_BOOL = bool
cdef type PY_TYPE_CURSOR
cdef type PY_TYPE_DATE = datetime.date
cdef type PY_TYPE_DATETIME = datetime.datetime
cdef type PY_TYPE_DECIMAL = decimal.Decimal
cdef type PY_TYPE_DB_OBJECT
cdef type PY_TYPE_DB_OBJECT_TYPE
cdef type PY_TYPE_JSON_ID
cdef type PY_TYPE_INTERVAL_YM
cdef type PY_TYPE_LOB
cdef type PY_TYPE_TIMEDELTA = datetime.timedelta
cdef type PY_TYPE_VAR
cdef type PY_TYPE_FETCHINFO

cdef const char* DRIVER_NAME = "python-oracledb"
cdef const char* DRIVER_VERSION
cdef const char* DRIVER_INSTALLATION_URL = \
        "https://python-oracledb.readthedocs.io/en/" \
        "latest/user_guide/initialization.html"
cdef const char* ENCODING_UTF8 = "UTF-8"
cdef const char* ENCODING_UTF16 = "UTF-16BE"

cdef int get_preferred_num_type(int16_t precision, int8_t scale):
    if scale == 0 or (scale == -127 and precision == 0):
        return NUM_TYPE_INT
    return NUM_TYPE_FLOAT


include "impl/base/constants.pxi"
include "impl/base/defaults.pyx"
include "impl/base/utils.pyx"
include "impl/base/pipeline.pyx"
include "impl/base/buffer.pyx"
include "impl/base/parsers.pyx"
include "impl/base/oson.pyx"
include "impl/base/vector.pyx"
include "impl/base/connect_params.pyx"
include "impl/base/pool_params.pyx"
include "impl/base/connection.pyx"
include "impl/base/pool.pyx"
include "impl/base/cursor.pyx"
include "impl/base/var.pyx"
include "impl/base/bind_var.pyx"
include "impl/base/dbobject.pyx"
include "impl/base/lob.pyx"
include "impl/base/soda.pyx"
include "impl/base/queue.pyx"
include "impl/base/subscr.pyx"



#------------------------------------------------------------------------------
# Copyright (c) 2022, 2024, Oracle and/or its affiliates.
#
# This software is dual-licensed to you under the Universal Permissive License
# (UPL) 1.0 as shown at https://oss.oracle.com/licenses/upl and Apache License
# 2.0 as shown at http://www.apache.org/licenses/LICENSE-2.0. You may choose
# either license.
#
# If you elect to accept the software under the Apache License, Version 2.0,
# the following applies:
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#------------------------------------------------------------------------------

#------------------------------------------------------------------------------
# utils.pyx
#
# Cython file defining utility methods (embedded in base_impl.pyx).
#------------------------------------------------------------------------------

cdef int _set_bool_param(dict args, str name, bint *out_val) except -1:
    """
    Sets a boolean parameter to the value provided in the dictionary. This can
    be a case-insenstive string matching on/off, yes/no or true/false (such as
    when parsed from a connect string). It can also be a directly passed
    argument which will be explicitly converted to a boolean value.
    """
    in_val = args.get(name)
    if in_val is not None:
        if isinstance(in_val, str):
            out_val[0] = (in_val.strip().lower() in ("on", "yes", "true"))
        else:
            out_val[0] = bool(in_val)


cdef int _set_duration_param(dict args, str name, double *out_val) except -1:
    """
    Sets a duration parameter to the value provided in the dictionary. This can
    be a string (such as when parsed from a connect string) containing a
    floating point value followed by an otional unit: ms (milliseconds), sec
    (seconds) or min (minutes). It can also be a directly passed argument which
    will be explicitly converted to a floating point value.
    """
    in_val = args.get(name)
    if in_val is not None:
        if isinstance(in_val, str):
            in_val = in_val.strip().lower()
            if in_val.endswith("sec"):
                out_val[0] = float(in_val[:-3].strip())
            elif in_val.endswith("ms"):
                out_val[0] = float(in_val[:-2].strip()) / 1000
            elif in_val.endswith("min"):
                out_val[0] = float(in_val[:-3].strip()) * 60
            else:
                out_val[0] = float(in_val.strip())
        else:
            out_val[0] = float(in_val)


cdef int _set_int_param(dict args, str name, int* out_val) except -1:
    """
    Sets an integer parameter to the value provided in the dictionary. This
    can be a string (such as when parsed from a connect string). It can also be
    a directly passed argument which will be explicitly converted to an integer
    value.
    """
    in_val = args.get(name)
    if in_val is not None:
        out_val[0] = int(in_val)


cdef int _set_uint_param(dict args, str name, uint32_t* out_val) except -1:
    """
    Sets an unsigned integer parameter to the value provided in the dictionary.
    This can be a string (such as when parsed from a connect string). It can
    also be a directly passed argument which will be explicitly converted to an
    integer value.
    """
    in_val = args.get(name)
    if in_val is not None:
        out_val[0] = int(in_val)


cdef int _set_obj_param(dict args, str name, object target) except -1:
    """
    Sets an object parameter to the value provided in the dictionary, if a
    value is provided. This value is then set directly on the target.
    """
    in_val = args.get(name)
    if in_val is not None:
        setattr(target, name, in_val)


cdef int _set_purity_param(dict args, str name, uint32_t* out_val) except -1:
    """
    Sets a purity parameter to the value provided in the dictionary. This
    must be one of "new" or "self" currently (or the equivalent constants, if
    specified directly). If it is not one of these values an error is raised.
    """
    cdef bint ok = True
    in_val = args.get(name)
    if in_val is not None:
        if isinstance(in_val, str):
            in_val = in_val.lower()
            if in_val == "new":
                out_val[0] = PURITY_NEW
            elif in_val == "self":
                out_val[0] = PURITY_SELF
            else:
                ok = False
        elif isinstance(in_val, int):
            if in_val == PURITY_NEW:
                out_val[0] = PURITY_NEW
            elif in_val == PURITY_SELF:
                out_val[0] = PURITY_SELF
            elif in_val == PURITY_DEFAULT:
                out_val[0] = PURITY_DEFAULT
            else:
                ok = False
        else:
            ok = False
        if not ok:
            errors._raise_err(errors.ERR_INVALID_POOL_PURITY,
                              purity=in_val)


cdef int _set_ssl_version_param(dict args, str name, object target) except -1:
    """
    Sets a SSL version parameter to the value specified. This must be one of
    the values "tlsv1.2" or "tlsv1.3". If it is not one of these values
    an error is raised. If a value is specified and meets the criteria it is
    set directly on the target (since strings are treated as Python objects).
    """
    in_val = args.get(name)
    if in_val is not None:
        if isinstance(in_val, str):
            in_val = in_val.lower()
            if in_val == "tlsv1.2":
                in_val = ssl.TLSVersion.TLSv1_2
            elif in_val == "tlsv1.3":
                in_val = ssl.TLSVersion.TLSv1_3
        if in_val is not ssl.TLSVersion.TLSv1_2 \
                and in_val is not ssl.TLSVersion.TLSv1_3:
            errors._raise_err(errors.ERR_INVALID_SSL_VERSION,
                              ssl_version=in_val)
        setattr(target, name, in_val)


cdef int _set_str_param(dict args, str name, object target) except -1:
    """
    Sets a string parameter to the value provided in the dictionary. If a value
    is specified it is set directly on the target (since strings are treated as
    Python objects).
    """
    in_val = args.get(name)
    if in_val:
        setattr(target, name, str(in_val))


def init_base_impl(package):
    """
    Initializes globals after the package has been completely initialized. This
    is to avoid circular imports and eliminate the need for global lookups.
    """
    global \
        errors, \
        exceptions, \
        utils, \
        DRIVER_VERSION, \
        PY_TYPE_ASYNC_CURSOR, \
        PY_TYPE_ASYNC_LOB, \
        PY_TYPE_CURSOR, \
        PY_TYPE_DB_OBJECT, \
        PY_TYPE_DB_OBJECT_TYPE, \
        PY_TYPE_FETCHINFO, \
        PY_TYPE_INTERVAL_YM, \
        PY_TYPE_JSON_ID, \
        PY_TYPE_LOB, \
        PY_TYPE_VAR

    errors = package.errors
    exceptions = package.exceptions
    utils = package.utils
    DRIVER_VERSION = package.__version__
    PY_TYPE_ASYNC_CURSOR = package.AsyncCursor
    PY_TYPE_ASYNC_LOB = package.AsyncLOB
    PY_TYPE_CURSOR = package.Cursor
    PY_TYPE_DB_OBJECT = package.DbObject
    PY_TYPE_DB_OBJECT_TYPE = package.DbObjectType
    PY_TYPE_FETCHINFO = package.FetchInfo
    PY_TYPE_INTERVAL_YM = package.IntervalYM
    PY_TYPE_JSON_ID = package.JsonId
    PY_TYPE_LOB = package.LOB
    PY_TYPE_VAR = package.Var



# -----------------------------------------------------------------------------
# Copyright (c) 2023, 2024, Oracle and/or its affiliates.
#
# This software is dual-licensed to you under the Universal Permissive License
# (UPL) 1.0 as shown at https://oss.oracle.com/licenses/upl and Apache License
# 2.0 as shown at http://www.apache.org/licenses/LICENSE-2.0. You may choose
# either license.
#
# If you elect to accept the software under the Apache License, Version 2.0,
# the following applies:
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# -----------------------------------------------------------------------------

"""
6800 - Module for testing error objects with asyncio
"""

import pickle
import unittest

import oracledb
import test_env


@unittest.skipUnless(
    test_env.get_is_thin(), "asyncio not supported in thick mode"
)
class TestCase(test_env.BaseAsyncTestCase):
    async def test_6800(self):
        "6800 - test parse error returns offset correctly"
        with self.assertRaises(oracledb.Error) as cm:
            await self.cursor.execute("begin t_Missing := 5; end;")
        (error_obj,) = cm.exception.args
        self.assertEqual(error_obj.full_code, "ORA-06550")
        self.assertEqual(error_obj.offset, 6)

    async def test_6801(self):
        "6801 - test picking/unpickling an error object"
        with self.assertRaises(oracledb.Error) as cm:
            await self.cursor.execute(
                """
                begin
                    raise_application_error(-20101, 'Test!');
                end;
                """
            )
        (error_obj,) = cm.exception.args
        self.assertIsInstance(error_obj, oracledb._Error)
        self.assertIn("Test!", error_obj.message)
        self.assertEqual(error_obj.code, 20101)
        self.assertEqual(error_obj.offset, 0)
        self.assertIsInstance(error_obj.isrecoverable, bool)
        new_error_obj = pickle.loads(pickle.dumps(error_obj))
        self.assertIsInstance(new_error_obj, oracledb._Error)
        self.assertEqual(new_error_obj.message, error_obj.message)
        self.assertEqual(new_error_obj.code, error_obj.code)
        self.assertEqual(new_error_obj.offset, error_obj.offset)
        self.assertEqual(new_error_obj.context, error_obj.context)
        self.assertEqual(new_error_obj.isrecoverable, error_obj.isrecoverable)

    async def test_6802(self):
        "6802 - test generation of full_code for ORA, DPI and DPY errors"
        cursor = self.conn.cursor()
        with self.assertRaises(oracledb.Error) as cm:
            await cursor.execute(None)
        (error_obj,) = cm.exception.args
        self.assertEqual(error_obj.full_code, "DPY-2001")

    async def test_6803(self):
        "6803 - test generation of error help portal URL"
        cursor = self.conn.cursor()
        with self.assertRaises(oracledb.Error) as cm:
            await cursor.execute("select 1 / 0 from dual")
        (error_obj,) = cm.exception.args
        to_check = "Help: https://docs.oracle.com/error-help/db/ora-01476/"
        self.assertIn(to_check, error_obj.message)

    async def test_6804(self):
        "6804 - verify warning is generated when creating a procedure"
        proc_name = "bad_proc_1704"
        self.assertIsNone(self.cursor.warning)
        await self.cursor.execute(
            f"""
            create or replace procedure {proc_name} as
            begin
                null
            end;
            """
        )
        self.assertEqual(self.cursor.warning.full_code, "DPY-7000")
        await self.cursor.execute(
            f"""
            create or replace procedure {proc_name} as
            begin
                null;
            end;
            """
        )
        self.assertIsNone(self.cursor.warning)
        await self.cursor.execute(f"drop procedure {proc_name}")

    async def test_6805(self):
        "6805 - verify warning is generated when creating a function"
        func_name = "bad_func_1705"
        await self.cursor.execute(
            f"""
            create or replace function {func_name}
            return number as
            begin
                return null
            end;
            """
        )
        self.assertEqual(self.cursor.warning.full_code, "DPY-7000")
        await self.cursor.execute(f"drop function {func_name}")
        self.assertIsNone(self.cursor.warning)

    async def test_6806(self):
        "6806 - verify warning is generated when creating a type"
        type_name = "bad_type_1706"
        await self.cursor.execute(
            f"""
            create or replace type {type_name} as object (
                x bad_type
            );
            """
        )
        self.assertEqual(self.cursor.warning.full_code, "DPY-7000")
        await self.cursor.execute(f"drop type {type_name}")
        self.assertIsNone(self.cursor.warning)


if __name__ == "__main__":
    test_env.run_test_cases()



#------------------------------------------------------------------------------
# Copyright (c) 2024, Oracle and/or its affiliates.
#
# This software is dual-licensed to you under the Universal Permissive License
# (UPL) 1.0 as shown at https://oss.oracle.com/licenses/upl and Apache License
# 2.0 as shown at http://www.apache.org/licenses/LICENSE-2.0. You may choose
# either license.
#
# If you elect to accept the software under the Apache License, Version 2.0,
# the following applies:
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#------------------------------------------------------------------------------

#------------------------------------------------------------------------------
# pipeline.pyx
#
# Cython file defining the PipelineImpl, PipelineOpImpl and
# PipelineOpResultImpl classes (embedded in base_impl.pyx).
#------------------------------------------------------------------------------

cdef class PipelineImpl:

    def __init__(self):
        self.operations = []


cdef class PipelineOpImpl:

    def __init__(
        self,
        uint8_t op_type,
        str statement = None,
        str name = None,
        object parameters = None,
        object keyword_parameters = None,
        object return_type = None,
        object rowfactory = None,
        uint32_t arraysize = 0,
        uint32_t num_rows = 0,
    ):
        self.op_type = op_type
        self.statement = statement
        self.name = name
        self.parameters = parameters
        self.keyword_parameters = keyword_parameters
        self.return_type = return_type
        self.rowfactory = rowfactory
        self.arraysize = arraysize
        self.num_rows = num_rows


cdef class PipelineOpResultImpl:

    def __init__(self, PipelineOpImpl op):
        self.operation = op

    cdef int _capture_err(self, Exception exc) except -1:
        """
        Captures the error in the result. If the error is not an error
        generated by the driver itself, wrap it so that the value in the error
        attribute is always an instance of the _Error class.
        """
        if isinstance(exc, exceptions.Error):
            self.error = exc.args[0]
        else:
            self.error = errors._create_err(
                errors.ERR_UNEXPECTED_PIPELINE_FAILURE, cause=exc
            )



# -----------------------------------------------------------------------------
# Copyright (c) 2021, 2024, Oracle and/or its affiliates.
#
# This software is dual-licensed to you under the Universal Permissive License
# (UPL) 1.0 as shown at https://oss.oracle.com/licenses/upl and Apache License
# 2.0 as shown at http://www.apache.org/licenses/LICENSE-2.0. You may choose
# either license.
#
# If you elect to accept the software under the Apache License, Version 2.0,
# the following applies:
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# -----------------------------------------------------------------------------

"""
3800 - Module for testing the input and output type handlers.
"""

import datetime
import json
import unittest

import oracledb
import test_env


class Building:
    def __init__(self, building_id, description, num_floors):
        self.building_id = building_id
        self.description = description
        self.num_floors = num_floors

    def __repr__(self):
        return f"<Building {self.building_id}: {self.description}>"

    def __eq__(self, other):
        if isinstance(other, Building):
            return (
                other.building_id == self.building_id
                and other.description == self.description
                and other.num_floors == self.num_floors
            )
        return NotImplemented

    def to_json(self):
        return json.dumps(self.__dict__)

    @classmethod
    def from_json(cls, value):
        result = json.loads(value)
        return cls(**result)


class TestCase(test_env.BaseTestCase):
    def building_in_converter(self, value):
        return value.to_json()

    def input_type_handler(self, cursor, value, num_elements):
        if isinstance(value, Building):
            return cursor.var(
                oracledb.STRING,
                arraysize=num_elements,
                inconverter=self.building_in_converter,
            )

    def output_type_handler(self, cursor, metadata):
        if metadata.type_code is oracledb.DB_TYPE_VARCHAR:
            return cursor.var(
                metadata.type_code,
                arraysize=cursor.arraysize,
                outconverter=Building.from_json,
            )

    def test_3800(self):
        "3800 - binding unsupported python object without input type handler"
        self.cursor.execute("truncate table TestTempTable")
        sql = "insert into TestTempTable (IntCol, StringCol1) values (:1, :2)"
        building = Building(1, "The First Building", 5)
        with self.assertRaisesFullCode("DPY-3002"):
            self.cursor.execute(sql, [building.building_id, building])

    def test_3801(self):
        "3801 - not callable input type handler"
        self.cursor.execute("truncate table TestTempTable")
        building = Building(1, "The First Building", 5)
        sql = "insert into TestTempTable (IntCol, StringCol1) values (:1, :2)"
        self.cursor.inputtypehandler = 5
        self.assertEqual(self.cursor.inputtypehandler, 5)
        with self.assertRaises(TypeError):
            self.cursor.execute(sql, (building.building_id, building))

    def test_3802(self):
        "3802 - binding unsupported python object with input type handler"
        self.cursor.execute("truncate table TestTempTable")
        building = Building(1, "The First Building", 5)
        self.cursor.inputtypehandler = self.input_type_handler
        self.cursor.execute(
            """
            insert into TestTempTable (IntCol, StringCol1)
            values (:1, :2)
            """,
            [building.building_id, building],
        )
        self.assertEqual(
            self.cursor.bindvars[1].inconverter, self.building_in_converter
        )
        self.conn.commit()
        self.cursor.execute("select IntCol, StringCol1 from TestTempTable")
        self.assertEqual(
            self.cursor.fetchall(),
            [(building.building_id, building.to_json())],
        )

    def test_3803(self):
        "3803 - input type handler and output type handler on cursor level"
        self.cursor.execute("truncate table TestTempTable")
        building_one = Building(1, "The First Building", 5)
        building_two = Building(2, "The Second Building", 87)
        sql = "insert into TestTempTable (IntCol, StringCol1) values (:1, :2)"
        cursor_one = self.conn.cursor()
        cursor_two = self.conn.cursor()
        cursor_one.inputtypehandler = self.input_type_handler
        cursor_one.execute(sql, [building_one.building_id, building_one])
        self.conn.commit()

        cursor_one.execute("select IntCol, StringCol1 from TestTempTable")
        self.assertEqual(
            cursor_one.fetchall(),
            [(building_one.building_id, building_one.to_json())],
        )
        with self.assertRaisesFullCode("DPY-3002"):
            cursor_two.execute(
                sql,
                (building_two.building_id, building_two),
            )

        cursor_two.outputtypehandler = self.output_type_handler
        cursor_two.execute("select IntCol, StringCol1 from TestTempTable")
        self.assertEqual(
            cursor_two.fetchall(), [(building_one.building_id, building_one)]
        )

    def test_3804(self):
        "3804 - input type handler and output type handler on connection level"
        self.cursor.execute("truncate table TestTempTable")
        building_one = Building(1, "The First Building", 5)
        building_two = Building(2, "The Second Building", 87)
        sql = "insert into TestTempTable (IntCol, StringCol1) values (:1, :2)"
        conn = test_env.get_connection()
        conn.inputtypehandler = self.input_type_handler
        self.assertEqual(conn.inputtypehandler, self.input_type_handler)

        cursor_one = conn.cursor()
        cursor_two = conn.cursor()
        cursor_one.execute(sql, [building_one.building_id, building_one])
        cursor_two.execute(sql, [building_two.building_id, building_two])
        conn.commit()

        expected_data = [
            (building_one.building_id, building_one),
            (building_two.building_id, building_two),
        ]
        conn.outputtypehandler = self.output_type_handler
        self.assertEqual(conn.outputtypehandler, self.output_type_handler)
        cursor_one.execute("select IntCol, StringCol1 from TestTempTable")
        self.assertEqual(
            cursor_one.fetchvars[1].outconverter, Building.from_json
        )
        self.assertEqual(cursor_one.fetchall(), expected_data)

        cursor_two.execute("select IntCol, StringCol1 from TestTempTable")
        self.assertEqual(cursor_two.fetchall(), expected_data)
        other_cursor = self.conn.cursor()
        with self.assertRaisesFullCode("DPY-3002"):
            other_cursor.execute(sql, (building_one.building_id, building_one))

    def test_3805(self):
        "3805 - output type handler with outconvert and null values"
        self.cursor.execute("truncate table TestTempTable")
        data_to_insert = [(1, "String 1"), (2, None), (3, "String 2")]
        self.cursor.executemany(
            """
            insert into TestTempTable (IntCol, StringCol1)
            values (:1, :2)
            """,
            data_to_insert,
        )
        self.conn.commit()

        def converter(value):
            return "CONVERTED"

        def output_type_handler(cursor, metadata):
            if metadata.type_code is oracledb.DB_TYPE_VARCHAR:
                return cursor.var(
                    str, outconverter=converter, arraysize=cursor.arraysize
                )

        self.cursor.outputtypehandler = output_type_handler
        self.cursor.execute(
            """
            select IntCol, StringCol1
            from TestTempTable
            order by IntCol
            """
        )
        expected_data = [(1, "CONVERTED"), (2, None), (3, "CONVERTED")]
        self.assertEqual(self.cursor.fetchall(), expected_data)

    @unittest.skipUnless(
        test_env.get_server_version() >= (21, 0), "unsupported server"
    )
    def test_3806(self):
        "3806 - output type handler for fetching 21c JSON"

        def output_type_handler(cursor, metadata):
            # fetch 21c JSON datatype when using python-oracledb thin mode
            if metadata.type_code is oracledb.DB_TYPE_JSON:
                return cursor.var(
                    str, arraysize=cursor.arraysize, outconverter=json.loads
                )
            # if using Oracle Client version < 21, then database returns BLOB
            # data type instead of JSON data type
            elif metadata.type_code is oracledb.DB_TYPE_BLOB:
                return cursor.var(
                    metadata.type,
                    arraysize=cursor.arraysize,
                    outconverter=lambda v: json.loads(v.read()),
                )

        self.cursor.execute("delete from TestJson")
        insert_sql = "insert into TestJson values (:1, :2)"
        json_data = [
            dict(name="John", city="Delhi"),
            dict(name="George", city="Bangalore"),
            dict(name="Sam", city="Mumbai"),
        ]
        data_to_insert = list(enumerate(json_data))
        json_as_string = self.conn.thin or test_env.get_client_version() < (
            21,
            0,
        )
        if json_as_string:
            # insert data as JSON string
            json_string_data = [(i, json.dumps(j)) for i, j in data_to_insert]
            self.cursor.executemany(insert_sql, json_string_data)
        else:
            # take advantage of direct binding
            self.cursor.setinputsizes(None, oracledb.DB_TYPE_JSON)
            self.cursor.executemany(insert_sql, data_to_insert)

        if json_as_string:
            self.cursor.outputtypehandler = output_type_handler
        self.cursor.execute("select * from TestJson")
        self.assertEqual(self.cursor.fetchall(), data_to_insert)

    def test_3807(self):
        "3807 - output type handler for encoding errors"

        if test_env.get_charset() != "AL32UTF8":
            self.skipTest("Database character set must be AL32UTF8")

        def output_type_handler(cursor, metadata):
            if metadata.type_code is oracledb.DB_TYPE_VARCHAR:
                return cursor.var(
                    metadata.type_code,
                    arraysize=cursor.arraysize,
                    encoding_errors="replace",
                )

        self.cursor.outputtypehandler = output_type_handler
        self.cursor.execute(
            "select utl_raw.cast_to_varchar2('41ab42cd43ef') from dual"
        )
        (result,) = self.cursor.fetchone()
        rc = chr(0xFFFD)
        expected_result = f"A{rc}B{rc}C{rc}"
        self.assertEqual(result, expected_result)

    def test_3808(self):
        "3808 - output type handler with object implementing __call__()"

        class TimestampOutputTypeHandler:

            def __init__(self, unit="s"):
                if unit == "ms":
                    self.factor = 1000
                else:
                    self.factor = 1

            def converter(self, d):
                return int(d.timestamp() * self.factor)

            def __call__(self, cursor, metadata):
                if metadata.type_code is oracledb.DB_TYPE_TIMESTAMP:
                    return cursor.var(
                        metadata.type_code,
                        arraysize=cursor.arraysize,
                        outconverter=self.converter,
                    )

        d = datetime.datetime.today()
        with self.conn.cursor() as cursor:
            cursor.outputtypehandler = TimestampOutputTypeHandler("ms")
            cursor.setinputsizes(oracledb.DB_TYPE_TIMESTAMP)
            cursor.execute("select :d from dual", [d])
            (result,) = cursor.fetchone()
            self.assertEqual(result, int(d.timestamp() * 1000))
        with self.conn.cursor() as cursor:
            cursor.outputtypehandler = TimestampOutputTypeHandler("s")
            cursor.setinputsizes(oracledb.DB_TYPE_TIMESTAMP)
            cursor.execute("select :d from dual", [d])
            (result,) = cursor.fetchone()
            self.assertEqual(result, int(d.timestamp()))


if __name__ == "__main__":
    test_env.run_test_cases()
